import time
import weakref
from abc import ABC, abstractmethod
from collections import OrderedDict, defaultdict, deque
from copy import deepcopy
from queue import Empty, Queue
from threading import Barrier, BrokenBarrierError, Event, Lock, Thread, Timer
from time import sleep

from eventsourcing.application.popo import PopoApplication
from eventsourcing.application.process import ProcessApplication, Prompt
from eventsourcing.application.simple import ApplicationWithConcreteInfrastructure
from eventsourcing.domain.model.decorators import retry
from eventsourcing.domain.model.events import subscribe, unsubscribe
from eventsourcing.exceptions import CausalDependencyFailed, EventSourcingError, \
    OperationalError, ProgrammingError, RecordConflictError

DEFAULT_POLL_INTERVAL = 5


class System(object):
    """
    A system object has a set of pipeline expressions, which involve
    process application classes. A system object can be run using
    a system runner.
    """
    def __init__(self, *pipeline_exprs, **kwargs):
        """
        Initialises a "process network" system object.

        :param pipeline_exprs: Pipeline expressions involving process application classes.

        Each pipeline expression of process classes shows directly which process
        follows which other process in the system.

        For example, the pipeline expression (A | B | C) shows that B follows A,
        and C follows B.

        The pipeline expression (A | A) shows that A follows A.

        The pipeline expression (A | B | A) shows that B follows A, and A follows B.

        The pipeline expressions ((A | B | A), (A | C | A)) are equivalent to (A | B | A | C | A).
        """
        self.pipelines_exprs = pipeline_exprs
        self.setup_tables = kwargs.get('setup_tables', False)
        self.infrastructure_class = kwargs.get('infrastructure_class', None)
        self.use_direct_query_if_available = kwargs.get('use_direct_query_if_available', False)

        self.session = kwargs.get('session', None)

        self.process_classes = OrderedDict()
        for pipeline_expr in self.pipelines_exprs:
            for process_class in pipeline_expr:
                process_name = process_class.__name__.lower()
                if process_name not in self.process_classes:
                    self.process_classes[process_name] = process_class

        self.processes = {}
        self.is_session_shared = True
        self.shared_session = None

        # Determine which process follows which.
        self.followers = OrderedDict()
        # A following is a list of process classes followed by a process class.
        # Todo: Factor this out, it's confusing. (Only used in ActorModelRunner now).
        self.followings = OrderedDict()
        for pipeline_expr in self.pipelines_exprs:
            previous_name = None
            for process_class in pipeline_expr:
                process_name = process_class.__name__.lower()
                try:
                    follows = self.followings[process_name]
                except KeyError:
                    follows = []
                    self.followings[process_name] = follows

                try:
                    self.followers[process_name]
                except KeyError:
                    self.followers[process_name] = []

                if previous_name and previous_name not in follows:
                    follows.append(previous_name)
                    followers = self.followers[previous_name]
                    followers.append(process_name)

                previous_name = process_name

    def construct_app(self, process_class, infrastructure_class=None, **kwargs):
        """
        Constructs process application from given ``process_class``.
        """
        kwargs = dict(kwargs)
        if 'session' not in kwargs and process_class.is_constructed_with_session:
            kwargs['session'] = self.session or self.shared_session
        if 'setup_tables' not in kwargs and self.setup_tables:
            kwargs['setup_table'] = self.setup_tables

        if not isinstance(process_class, ApplicationWithConcreteInfrastructure):

            # If process class isn't already an infrastructure class, then
            # use given arg, or attribute of this object, or PopoApplication.
            if infrastructure_class is None:
                infrastructure_class = self.infrastructure_class or PopoApplication

            if not issubclass(infrastructure_class, ApplicationWithConcreteInfrastructure):
                raise ProgrammingError(
                    'Given infrastructure_class {} is not subclass of {}'
                    ''.format(infrastructure_class, ApplicationWithConcreteInfrastructure)
                )

            # Subclass the process application class with the infrastructure class.
            process_class = process_class.mixin(infrastructure_class)

        # Construct the process application.
        process = process_class(**kwargs)

        # Catch the session, so it can be shared.
        if self.session is None:
            if process_class.is_constructed_with_session and self.is_session_shared:
                if self.shared_session is None:
                    self.shared_session = process.session

        return process

    def is_prompt(self, event):
        """
        Predicate for subscribing to publication of Prompt objects.
        """
        return isinstance(event, Prompt)

    def __enter__(self):
        """
        Supports usage of a system object as a context manager.
        """
        runner = SingleThreadedRunner(
            system=self,
            use_direct_query_if_available=self.use_direct_query_if_available,

        )
        self.__runner = weakref.ref(runner)
        self.__runner().__enter__()
        return runner

    def __exit__(self, exc_type, exc_val, exc_tb):
        """
        Supports usage of a system object as a context manager.
        """
        self.__runner().__exit__(exc_type, exc_val, exc_tb)
        del (self.__runner)

    def __getattr__(self, process_name, infrastructure_class=None):
        """
        Supports accessing process application by name as object attribute.

        :param process_name:
        :param infrastructure_class:
        :return: A process application object.
        :rtype: ProcessApplication
        """
        if self.processes and process_name in self.processes:
            # Todo: Move to having process application instances on runner only.
            process = self.processes[process_name]
        else:
            try:
                process_class = self.process_classes[process_name]
            except KeyError:
                raise AttributeError(process_name)
            else:
                process = self.construct_app(
                    process_class=process_class,
                    setup_table=self.setup_tables,
                    infrastructure_class=infrastructure_class,
                )
                # Todo: Move to having process application instances on runner only.
                self.processes[process_name] = process
        return process

    def bind(self, infrastructure_class):
        """
        Constructs a system object that has an infrastructure class
        from system object constructed without infrastructure class.

        Raises ProgrammingError if already have an infrastructure class.

        :param infrastructure_class:
        :return: System object that has an infrastructure class.
        :rtype: System
        """
        # Check system doesn't already have an infrastructure class.
        if self.infrastructure_class:
            raise ProgrammingError('System already has an infrastructure class')

        # Clone the system object, and set the infrastructure class.
        system = object.__new__(type(self))
        system.__dict__.update(dict(deepcopy(self.__dict__)))
        system.__dict__.update(infrastructure_class=infrastructure_class)
        return system


class SystemRunner(ABC):

    def __init__(self, system: System, infrastructure_class=None, setup_tables=False,
                 use_direct_query_if_available=False):
        self.system = system
        self.infrastructure_class = infrastructure_class or self.system.infrastructure_class
        self.setup_tables = setup_tables
        self.use_direct_query_if_available = use_direct_query_if_available or \
                                             system.use_direct_query_if_available

        # Todo: Move to having process application instances on runner only.
        self.processes = self.system.processes

    def __enter__(self):
        """
        Supports usage of a system runner as a context manager.
        """
        self.start()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        """
        Supports usage of a system runner as a context manager.
        """
        self.close()

    @abstractmethod
    def start(self):
        """
        Starts running the system.

        Abstract method which must be implemented on concrete descendants.
        """

    def close(self):
        """
        Closes a running system.
        """
        self.system.shared_session = None
        if self.system.processes:
            for process_name, process in self.system.processes.items():
                process.close()
            self.system.processes.clear()

    def __getattr__(self, process_name):
        """
        Supports accessing process application by name as object attribute.

        :param process_name:
        :param infrastructure_class:
        :return: A process application object.
        :rtype: ProcessApplication
        """

        # Todo: Move to having process application instances on runner only.
        return self.system.__getattr__(process_name, infrastructure_class=self.infrastructure_class)

# Todo: Support passing SQLAlchemy session into runner.

class InProcessRunner(SystemRunner):
    """
    Runs a system in the current process,
    either in the current thread, or with
    one thread for each process in the system.
    """

    def start(self):
        assert len(self.system.processes) == 0, "Already running"

        # Construct the processes.
        for process_class in self.system.process_classes.values():
            process = self.system.construct_app(
                process_class=process_class,
                infrastructure_class=self.infrastructure_class,
                setup_table=self.setup_tables or self.system.setup_tables,
                use_direct_query_if_available=self.use_direct_query_if_available,
            )
            self.system.processes[process.name] = process

        # Tell each process about the processes it follows.
        for followed_name, followers in self.system.followers.items():
            followed = self.system.processes[followed_name]
            followed_log = followed.notification_log
            for follower_name in followers:
                follower = self.system.processes[follower_name]
                follower.follow(followed_name, followed_log)

        # Do something to propagate prompts.
        subscribe(
            predicate=self.system.is_prompt,
            handler=self.handle_prompt,
        )

    @abstractmethod
    def handle_prompt(self, prompt):
        """
        Handles publication of a prompt.

        Abstract method which must be implemented on concrete descendants.
        """

    def close(self):
        super(InProcessRunner, self).close()

        unsubscribe(
            predicate=self.system.is_prompt,
            handler=self.handle_prompt,
        )


class SingleThreadedRunner(InProcessRunner):
    """
    Runs a system in the current thread.
    """

    def __init__(self, system: System, infrastructure_class=None, *args, **kwargs):
        super(SingleThreadedRunner, self).__init__(
            system=system, infrastructure_class=infrastructure_class, *args, **kwargs
        )
        self.pending_prompts = Queue()
        self.iteration_lock = Lock()

    def handle_prompt(self, prompt):
        self.run_followers(prompt)

    def run_followers(self, prompt):
        """
        First caller adds a prompt to queue and
        runs followers until there are no more
        pending prompts.

        Subsequent callers just add a prompt
        to the queue, avoiding recursion.
        """
        assert isinstance(prompt, Prompt)
        # Put the prompt on the queue.
        self.pending_prompts.put(prompt)

        if self.iteration_lock.acquire(False):
            start_time = time.time()
            i = 0
            try:
                while True:
                    try:
                        prompt = self.pending_prompts.get(False)
                    except Empty:
                        break
                    else:
                        followers = self.system.followers[prompt.process_name]
                        for follower_name in followers:
                            follower = self.system.processes[follower_name]
                            follower.run(prompt)
                            i += 1
                        self.pending_prompts.task_done()
            finally:
                run_frequency = i / (time.time() - start_time)
                # print(f"Run frequency: {run_frequency}")
                self.iteration_lock.release()

    # This is the old way of doing it, with recursion.
    # def run_followers_with_recursion(self, prompt):
    #     followers = self.system.followers[prompt.process_name]
    #     for follower_name in followers:
    #         follower = self.processes[follower_name]
    #         follower.run(prompt)
    #


class MultiThreadedRunner(InProcessRunner):
    """
    Runs a system with a thread for each process.
    """

    def __init__(self, system: System, poll_interval=None, clock_speed=None, **kwargs):
        super(MultiThreadedRunner, self).__init__(system=system, **kwargs)
        self.poll_interval = poll_interval or DEFAULT_POLL_INTERVAL
        assert isinstance(system, System)
        self.threads = {}
        self.clock_speed = clock_speed
        if self.clock_speed:
            self.clock_event = Event()
            self.stop_clock_event = Event()
        else:
            self.clock_event = None
            self.stop_clock_event = None

    def start(self):
        super(MultiThreadedRunner, self).start()
        assert not self.threads, "Already started"

        self.inboxes = {}
        self.outboxes = {}
        self.clock_events = []

        # Setup queues.
        for process_name, upstream_names in self.system.followings.items():
            inbox_id = process_name.lower()
            if inbox_id not in self.inboxes:
                self.inboxes[inbox_id] = Queue()
            for upstream_class_name in upstream_names:
                outbox_id = upstream_class_name.lower()
                if outbox_id not in self.outboxes:
                    self.outboxes[outbox_id] = PromptOutbox()
                if inbox_id not in self.outboxes[outbox_id].downstream_inboxes:
                    self.outboxes[outbox_id].downstream_inboxes[inbox_id] = self.inboxes[inbox_id]

        # Construct application threads.
        for process_name, process in self.system.processes.items():
            process_instance_id = process_name
            if self.clock_event:
                process.clock_event = self.clock_event
                process.tick_interval = 1 / self.clock_speed

            thread = PromptQueuedApplicationThread(
                process=process,
                poll_interval=self.poll_interval,
                inbox=self.inboxes[process_instance_id],
                outbox=self.outboxes[process_instance_id],
                # Todo: Is it better to clock the prompts or the notifications?
                # clock_event=clock_event
            )
            self.threads[process_instance_id] = thread

        # Start application threads.
        for thread in self.threads.values():
            thread.start()

        # Start clock.
        if self.clock_speed:
            self.start_clock()

    def start_clock(self):
        tick_interval = 1 / self.clock_speed
        # print(f"Tick interval: {tick_interval:.6f}s")
        self.last_tick = None
        self.this_tick = None
        self.tick_adjustment = 0

        def set_clock_event():
            if self.stop_clock_event.is_set():
                return

            self.this_tick = time.process_time()

            if self.last_tick:
                tick_size = self.this_tick - self.last_tick

                tick_oversize = tick_size - tick_interval
                tick_oversize_percentage = 100 * (tick_oversize) / tick_interval
                if tick_oversize_percentage > 300:
                    print(f"Warning: Tick over size: {tick_size :.6f}s {tick_oversize_percentage:.2f}%")

                if abs(tick_oversize_percentage) < 300:
                    self.tick_adjustment += 0.5 * tick_interval * tick_oversize
                    max_tick_adjustment = 0.5 * tick_interval
                    min_tick_adjustment = 0
                    self.tick_adjustment = min(self.tick_adjustment, max_tick_adjustment)
                    self.tick_adjustment = max(self.tick_adjustment, min_tick_adjustment)

            self.last_tick = self.this_tick

            self.clock_event.set()
            self.clock_event.clear()

            if not self.stop_clock_event.is_set():
                set_timer()

        def set_timer():
            # print(f"Tick adjustment: {self.tick_adjustment:.6f}")
            if self.last_tick is not None:
                time_since_last_tick = time.process_time() - self.last_tick
                time_remaining = tick_interval - time_since_last_tick
                timer_interval = time_remaining - self.tick_adjustment
                if timer_interval < 0:
                    timer_interval = 0
                    # print("Warning: clock thread is running flat out!")
            else:
                timer_interval = 0
            timer = Timer(timer_interval, set_clock_event)
            timer.start()

        set_timer()

    def handle_prompt(self, prompt):
        self.broadcast_prompt(prompt)

    def broadcast_prompt(self, prompt):
        outbox_id = prompt.process_name
        assert outbox_id in self.outboxes, (outbox_id, self.outboxes.keys())
        self.outboxes[outbox_id].put(prompt)

    def close(self):
        super(MultiThreadedRunner, self).close()

        if self.clock_event is not None:
            self.clock_event.set()

        if self.stop_clock_event is not None:
            self.stop_clock_event.set()

        for thread in self.threads.values():
            thread.inbox.put('QUIT')

        for thread in self.threads.values():
            thread.join(timeout=10)

        self.threads.clear()


class PromptQueuedApplicationThread(Thread):
    """
    Application thread which uses queues of prompts.

    It loops on an "inbox" queue of prompts, and
    adds its prompts to an "outbox" queue.
    """

    def __init__(self, process: ProcessApplication, poll_interval=DEFAULT_POLL_INTERVAL,
                 inbox=None, outbox=None, clock_event=None):
        super(PromptQueuedApplicationThread, self).__init__(daemon=True)
        self.process = process
        self.poll_interval = poll_interval
        self.inbox = inbox
        self.outbox = outbox
        self.clock_event = clock_event

    def run(self):
        self.loop_on_prompts()

    def loop_on_prompts(self):

        # Loop on getting prompts.
        while True:
            try:
                # Todo: Make the poll interval gradually increase if there are only timeouts?
                prompt = self.inbox.get(timeout=self.poll_interval)

            except Empty:
                # Basically, we're polling after a timeout.
                if self.clock_event is None:
                    self.run_process()

            else:
                self.inbox.task_done()

                if prompt == 'QUIT':
                    self.process.close()
                    break

                else:
                    if self.clock_event is not None:
                        self.clock_event.wait()
                        started = time.time()

                    self.run_process(prompt)

                    if self.clock_event is not None:
                        ended = time.time()
                        duration = ended - started
                        if self.clock_event.is_set():
                            print(f"Warning: Process {self.process.name} overran clock cycle: {duration}")
                        else:
                            print(f"Info: Process {self.process.name} ran within clock cycle: {duration}")

    def run_process(self, prompt=None):
        try:
            self._run_process(prompt)
        except CausalDependencyFailed:
            pass
        except EventSourcingError:
            pass

    @retry(CausalDependencyFailed, max_attempts=100, wait=0.2)
    @retry((OperationalError, RecordConflictError), max_attempts=100, wait=0.01)
    def _run_process(self, prompt=None):
        self.process.run(prompt)


class PromptOutbox(object):
    """
    Has a collection of downstream prompt inboxes.

    """

    def __init__(self):
        self.downstream_inboxes = {}

    def put(self, prompt):
        """
        Puts prompt in each downstream inbox (an actual queue).
        """
        for queue in self.downstream_inboxes.values():
            queue.put(prompt)


class SteppingRunner(InProcessRunner):

    def __init__(self, normal_speed=1, scale_factor=1, is_verbose=False, *args, **kwargs):
        super(SteppingRunner, self).__init__(*args, **kwargs)
        self.normal_speed = normal_speed
        self.scale_factor = scale_factor
        self.is_verbose = is_verbose
        if scale_factor:
            self.tick_interval = 1 / (normal_speed * scale_factor)
        else:
            self.tick_interval = 0
        if self.is_verbose:
            print(f"Tick interval: {self.tick_interval:.6f}s")
        self.clock_thread = None

    def call_in_future(self, cmd, ticks_delay):
        self.clock_thread.call_in_future(cmd, ticks_delay)


class SteppingSingleThreadedRunner(SteppingRunner):

    def __init__(self, *args, **kwargs):
        super(SteppingSingleThreadedRunner, self).__init__(*args, **kwargs)
        self.seen_prompt_events = {}
        self.stop_event = Event()

    def start(self):
        super(SteppingSingleThreadedRunner, self).start()

        self.clock_thread = ProcessRunningClockThread(
            normal_speed=self.normal_speed,
            scale_factor=self.scale_factor,
            stop_event=self.stop_event,
            is_verbose=self.is_verbose,
            seen_prompt_events=self.seen_prompt_events,
            processes=self.system.processes,
            use_direct_query_if_available=self.use_direct_query_if_available
        )
        self.clock_thread.start()

    def handle_prompt(self, prompt):
        """
        Ignores prompts.
        """

    def close(self):
        self.stop_event.set()
        super(SteppingSingleThreadedRunner, self).close()
        self.clock_thread.join(5)
        if self.clock_thread.isAlive():
            print("Warning: clock thread was still alive")


class ClockThread(Thread):
    def __init__(self, *args, **kwargs):
        super(ClockThread, self).__init__()
        self.future_cmds = defaultdict(list)
        self.tick_count = 0

    def call_in_future(self, cmd, ticks_delay):
        ticks_delay = max(ticks_delay, 1)
        self.future_cmds[ticks_delay + self.tick_count].append(cmd)

    def call_commands(self):
        for future_cmd in self.future_cmds.get(self.tick_count, []):
            future_cmd()


class ProcessRunningClockThread(ClockThread):
    def __init__(self, normal_speed, scale_factor, stop_event: Event,
                 is_verbose=False, seen_prompt_events=None, processes=None,
                 use_direct_query_if_available=False):
        super(ProcessRunningClockThread, self).__init__(daemon=True)
        self.normal_speed = normal_speed
        self.scale_factor = scale_factor
        self.stop_event = stop_event
        self.seen_prompt_events = seen_prompt_events
        self.processes = processes
        self.use_direct_query_if_available = use_direct_query_if_available
        self.last_tick_time = None
        self.last_process_time = None
        self.all_tick_durations = deque()
        self.tick_adjustment = 0.0
        self.is_verbose = is_verbose
        if normal_speed and scale_factor:
            self.tick_interval = 1 / normal_speed / scale_factor
        else:
            self.tick_interval = None
        if self.tick_interval:

            self.tick_durations_window_size = max(100, int(round(1 / self.tick_interval, 0)))
        else:
            self.tick_durations_window_size = 1000
        # Construct lists of followers for each process.
        self.followers = {}
        for process_name, process in self.processes.items():
            self.followers[process_name] = []
        for process_name, process in self.processes.items():
            for upstream_process_name in process.readers:
                self.followers[upstream_process_name].append(process_name)

        # Construct a notification log reader for each process.
        self.readers = {}
        for process_name, process in self.processes.items():
            reader = process.notification_log_reader_class(
                notification_log=process.notification_log,
                use_direct_query_if_available=self.use_direct_query_if_available
            )
            self.readers[process_name] = reader

    @property
    def actual_clock_speed(self):
        if self.all_tick_durations:
            # Todo: Might need to lock this, to avoid "RuntimeError: deque mutated during iteration".
            durations = list(self.all_tick_durations)
            return len(durations) / sum(durations)
        else:
            return 0

    def run(self):
        # Get new notifications once.

        # loop_counter = 0

        while not self.stop_event.is_set():
            # print("Loop count: {}".format(loop_counter))
            try:
                # Get all notifications.
                all_notifications = {}
                for process_name in self.processes:
                    # seen_prompt = self.seen_prompt_events[process_name]
                    # if seen_prompt.is_set():
                    #     seen_prompt.clear()
                    reader = self.readers[process_name]
                    notifications = reader.read_list()
                    all_notifications[process_name] = notifications

                # Process all notifications.
                all_events = {}
                for process_name, notifications in all_notifications.items():
                    events = []
                    for notification in notifications:
                        # print(process_name, notification)
                        process = self.processes[process_name]
                        # It's not the follower process, but the method does the same thing.
                        event = process.get_event_from_notification(notification)
                        notification_id = notification['id']
                        events.append((notification_id, event))
                    all_events[process_name] = events

                for process_name, events in all_events.items():
                    # print(f"Process: {process_name}")
                    for follower_name in self.followers[process_name]:
                        follower_process = self.processes[follower_name]
                        # print(f"Follower: {follower_name}")
                        for notification_id, event in events:
                            # print(f"Notification: {notification_id}, {event}")
                            follower_process.process_upstream_event(event, notification_id, process_name)
                # Call commands delayed until this clock tick.
                self.call_commands()

            except:
                if self.stop_event.is_set():
                    break
                else:
                    # print("Error... Loop count: {}".format(loop_counter))
                    self.stop_event.set()
                    raise
            else:
                tick_time = time.time()
                process_time = time.process_time()
                if self.last_tick_time is not None:
                    tick_duration = tick_time - self.last_tick_time
                    self.all_tick_durations.append(tick_duration)
                    if len(self.all_tick_durations) > self.tick_durations_window_size:
                        self.all_tick_durations.popleft()

                    if self.is_verbose:
                        process_duration = process_time - self.last_process_time
                        intensity = 100 * process_duration / tick_duration
                        clock_speed = 1 / tick_duration
                        real_time = self.tick_count / self.normal_speed
                        print(f"Tick {self.tick_count:4}: {real_time:4.2f}s  {tick_duration:.6f}s, "
                              f"{intensity:6.2f}%, {clock_speed:6.1f}Hz, "
                              f"{self.actual_clock_speed:6.1f}Hz, {self.tick_adjustment:.6f}s"
                              )

                    if self.tick_interval:
                        tick_oversize = tick_duration - self.tick_interval
                        tick_oversize_percentage = 100 * (tick_oversize) / self.tick_interval
                        # if tick_oversize_percentage > 300:
                        #     print(f"Warning: Tick over size: { tick_duration :.6f}s {tick_oversize_percentage:.2f}%")

                        if abs(tick_oversize_percentage) < 300:
                            # Weight falls from 1 as reciprocal of count, to tick interval.
                            # weight = max(1 / self.tick_count, min(.1, self.tick_interval))
                            weight = 1 / (1 + self.tick_count * self.tick_interval) ** 2
                            # print(f"Weight: {weight:.4f}")
                            self.tick_adjustment += weight * tick_oversize
                            max_tick_adjustment = 1.0 * self.tick_interval
                            min_tick_adjustment = 0
                            self.tick_adjustment = min(self.tick_adjustment, max_tick_adjustment)
                            self.tick_adjustment = max(self.tick_adjustment, min_tick_adjustment)

                self.last_tick_time = tick_time
                self.last_process_time = process_time
                self.tick_count += 1

                if self.tick_interval:
                    sleep_interval = self.tick_interval - self.tick_adjustment
                    sleep(max(sleep_interval, 0))


class SteppingMultiThreadedRunner(SteppingRunner):
    """
    Has a clock thread, and a thread for each application process
    in the system. The clock thread loops until stopped, waiting
    for a barrier, after sleeping for remaining tick interval timer.
    Application threads loop until stopped, waiting for the same
    barrier. Then, after all threads are waiting at the barrier,
    the barrier is lifted. The clock thread proceeds by sleeping
    for the clock tick interval. The application threads proceed by
    getting new notifications and processing all of them.

    There are actually two barriers, so that each application thread
    waits before getting notifications, and then waits for all processes
    to complete getting notification before processing the notifications
    through the application policy. This avoids events created by a process
    application "bleeding" into the notifications of another process
    application in the same clock cycle.


    Todo:
    Receive prompts, but set an event for the prompting process, to avoid unnecessary runs.

    Allow commands to be scheduled at future clock tick number, and execute when reached.

    """

    def __init__(self, *args, **kwargs):
        super(SteppingMultiThreadedRunner, self).__init__(*args, **kwargs)
        self.seen_prompt_events = {}
        self.fetch_barrier = None
        self.execute_barrier = None
        self.application_threads = {}
        self.clock_thread = None
        self.stop_event = Event()

    def handle_prompt(self, prompt):
        seen_prompt = self.seen_prompt_events[prompt.process_name]
        seen_prompt.set()

    def start(self):
        super(SteppingMultiThreadedRunner, self).start()
        parties = 1 + len(self.system.processes)
        self.fetch_barrier = Barrier(parties)
        self.execute_barrier = Barrier(parties)

        # Create an event for each process.
        for process_name in self.system.processes:
            self.seen_prompt_events[process_name] = Event()

        # Construct application threads.
        for process_name, process in self.system.processes.items():
            process_instance_id = process_name

            thread = BarrierControlledApplicationThread(
                process=process,
                fetch_barrier=self.fetch_barrier,
                execute_barrier=self.execute_barrier,
                stop_event=self.stop_event,
            )
            self.application_threads[process_instance_id] = thread

        # Start application threads.
        for thread in self.application_threads.values():
            thread.start()

        # Start clock thread.
        self.clock_thread = BarrierControllingClockThread(
            normal_speed=self.normal_speed,
            scale_factor=self.scale_factor,
            tick_interval=self.tick_interval,
            fetch_barrier=self.fetch_barrier,
            execute_barrier=self.execute_barrier,
            stop_event=self.stop_event,
            is_verbose=self.is_verbose,
        )
        self.clock_thread.start()

    def close(self):
        self.stop_event.set()
        super(SteppingMultiThreadedRunner, self).close()
        self.execute_barrier.abort()
        self.fetch_barrier.abort()

        for thread in self.application_threads.values():
            thread.join(timeout=1)
            if thread.isAlive():
                print(f"Warning: application thread '{thread.process.name}' was still alive: {thread.state}")

        self.application_threads.clear()

        self.clock_thread.join()
        if self.clock_thread.isAlive():
            print(f"Warning: clock thread was still alive")


class BarrierControlledApplicationThread(Thread):
    def __init__(self, process: ProcessApplication, fetch_barrier: Barrier,
                 execute_barrier: Barrier, stop_event: Event):
        super(BarrierControlledApplicationThread, self).__init__(daemon=True)
        self.process_application = process
        self.fetch_barrier = fetch_barrier
        self.execute_barrier = execute_barrier
        self.stop_event = stop_event

    def run(self):
        while not self.stop_event.is_set():
            # Isolate "fetch" and "execute" steps, to avoid
            # events created in one application being processed by
            # another application in the same tick. Race condition
            # where one process writes new events before another has
            # read all notifications from last tick. Actually, need
            # to get all notifications from all upstream applications
            # and then process the notifications. The run() method
            # gets notifications from one, then processes, then gets
            # from another, which makes the race condition probable.
            all_notifications = []
            try:
                self.fetch_barrier.wait()
            except BrokenBarrierError:
                self.abort()
            else:
                try:
                    # Get all notifications.
                    for upstream_name in self.process_application.readers:
                        notifications = list(self.process_application.read_reader(upstream_name))
                        all_notifications.append((upstream_name, notifications))

                except:
                    self.abort()
                    raise

            if self.stop_event.is_set():
                break

            try:
                self.execute_barrier.wait()
            except BrokenBarrierError:
                self.abort()
            else:
                try:
                    # Process all notifications.
                    for upstream_name, notifications in all_notifications:
                        for notification in notifications:
                            event = self.process_application.get_event_from_notification(notification)
                            self.process_application.process_upstream_event(event, notification['id'], upstream_name)
                except:
                    self.abort()
                    raise

            try:
                self.execute_barrier.wait()
            except BrokenBarrierError:
                self.abort()

    def abort(self):
        self.stop_event.set()
        self.fetch_barrier.abort()
        self.execute_barrier.abort()


class BarrierControllingClockThread(ClockThread):
    def __init__(self, normal_speed, scale_factor, tick_interval,
                 fetch_barrier: Barrier, execute_barrier: Barrier,
                 stop_event: Event, is_verbose=False):
        super(BarrierControllingClockThread, self).__init__(daemon=True)
        # Todo: Remove the redundancy here.
        self.normal_speed = normal_speed
        self.scale_factor = scale_factor
        self.tick_interval = tick_interval
        self.fetch_barrier = fetch_barrier
        self.execute_barrier = execute_barrier
        self.stop_event = stop_event
        self.last_tick_time = None
        self.last_process_time = None
        self.all_tick_durations = deque()
        self.tick_adjustment = 0.0
        self.is_verbose = is_verbose
        if self.tick_interval:

            self.tick_durations_window_size = max(1, int(round(1 / self.tick_interval, 0)))
        else:
            self.tick_durations_window_size = 100

    @property
    def actual_clock_speed(self):
        if self.all_tick_durations:
            durations = self.all_tick_durations
            return len(durations) / sum(durations)
        else:
            return 0

    def run(self):
        while not self.stop_event.is_set():

            try:
                self.fetch_barrier.wait()
                self.execute_barrier.wait()
                self.execute_barrier.wait()
                self.call_commands()
            except BrokenBarrierError:
                self.fetch_barrier.abort()
                self.execute_barrier.abort()
                self.stop_event.set()
            else:
                tick_time = time.time()
                process_time = time.process_time()
                if self.last_tick_time is not None:
                    tick_duration = tick_time - self.last_tick_time
                    self.all_tick_durations.append(tick_duration)
                    if len(self.all_tick_durations) > self.tick_durations_window_size:
                        self.all_tick_durations.popleft()

                    if self.is_verbose:
                        process_duration = process_time - self.last_process_time
                        intensity = 100 * process_duration / tick_duration
                        clock_speed = 1 / tick_duration
                        real_time = self.tick_count / self.normal_speed
                        print(f"Tick {self.tick_count:4}: {real_time:4.2f}s  {tick_duration:.6f}s, "
                              f"{intensity:6.2f}%, {clock_speed:6.1f}Hz, "
                              f"{self.actual_clock_speed:6.1f}Hz, {self.tick_adjustment:.6f}s"
                              )

                    if self.tick_interval:
                        tick_oversize = tick_duration - self.tick_interval
                        tick_oversize_percentage = 100 * (tick_oversize) / self.tick_interval
                        # if tick_oversize_percentage > 300:
                        #     print(f"Warning: Tick over size: { tick_duration :.6f}s {tick_oversize_percentage:.2f}%")

                        if abs(tick_oversize_percentage) < 300:
                            # Weight falls from 1 as reciprocal of count, to tick interval.
                            # weight = max(1 / self.tick_count, min(.1, self.tick_interval))
                            weight = 1 / (1 + self.tick_count * self.tick_interval) ** 2
                            # print(f"Weight: {weight:.4f}")
                            self.tick_adjustment += weight * tick_oversize
                            max_tick_adjustment = 1.0 * self.tick_interval
                            min_tick_adjustment = 0
                            self.tick_adjustment = min(self.tick_adjustment, max_tick_adjustment)
                            self.tick_adjustment = max(self.tick_adjustment, min_tick_adjustment)

                self.last_tick_time = tick_time
                self.last_process_time = process_time
                self.tick_count += 1

                if self.tick_interval:
                    sleep_interval = self.tick_interval - self.tick_adjustment
                    sleep(max(sleep_interval, 0))



from eventsourcing.application.simple import ApplicationWithConcreteInfrastructure
from eventsourcing.infrastructure.django.factory import DjangoInfrastructureFactory
from eventsourcing.infrastructure.django.utils import close_django_connection, setup_django


class DjangoApplication(ApplicationWithConcreteInfrastructure):
    infrastructure_factory_class = DjangoInfrastructureFactory

    def __init__(self, tracking_record_class=None, *args, **kwargs):
        self._tracking_record_class = tracking_record_class
        super(DjangoApplication, self).__init__(*args, **kwargs)

    @property
    def stored_event_record_class(cls):
        from eventsourcing.infrastructure.django.models import StoredEventRecord
        return StoredEventRecord

    @property
    def snapshot_record_class(cls):
        from eventsourcing.infrastructure.django.models import EntitySnapshotRecord
        return EntitySnapshotRecord

    @property
    def tracking_record_class(cls):
        from eventsourcing.infrastructure.django.models import NotificationTrackingRecord
        return NotificationTrackingRecord

    def construct_infrastructure(self, *args, **kwargs):
        tracking_record_class = self._tracking_record_class or self.tracking_record_class
        super(DjangoApplication, self).construct_infrastructure(
            tracking_record_class=tracking_record_class,
            *args, **kwargs
        )

    @classmethod
    def reset_connection_after_forking(cls):
        """
        Resets database connection after forking.
        """
        close_django_connection()
        setup_django()



import hashlib
import os

from eventsourcing.utils.transcoding import json_dumps

SALT_FOR_DATA_INTEGRITY = os.getenv('SALT_FOR_DATA_INTEGRITY', '')


def hash_object(json_encoder_class, obj):
    """
    Calculates SHA-256 hash of JSON encoded 'obj'.

    :param obj: Object to be hashed.
    :return: SHA-256 as hexadecimal string.
    :rtype str
    """
    s = json_dumps(
        (obj, SALT_FOR_DATA_INTEGRITY),
        cls=json_encoder_class,
    )
    return hashlib.sha256(s.encode()).hexdigest()



import os

from sqlalchemy import create_engine
from sqlalchemy.exc import InternalError
from sqlalchemy.orm import scoped_session, sessionmaker
from sqlalchemy.pool import StaticPool

from eventsourcing.infrastructure.datastore import AbstractDatastore, DatastoreSettings
from eventsourcing.infrastructure.sqlalchemy.records import Base

SQLITE_IN_MEMORY = 'sqlite:///:memory:'
DEFAULT_SQLALCHEMY_DB_URI = SQLITE_IN_MEMORY
# DEFAULT_SQLALCHEMY_DB_URI = 'sqlite:///FILE_SYSTEM_PATH'
# DEFAULT_SQLALCHEMY_DB_URI = 'mysql://username:password@localhost/eventsourcing'
# DEFAULT_SQLALCHEMY_DB_URI = 'postgresql://username:password@localhost:5432/eventsourcing'
DEFAULT_SQLALCHEMY_DB_POOL_SIZE = 5

class SQLAlchemySettings(DatastoreSettings):
    def __init__(self, uri=None, pool_size=None):
        self.uri = uri or os.getenv('DB_URI', DEFAULT_SQLALCHEMY_DB_URI)
        self.pool_size = int(pool_size or os.getenv('DB_POOL_SIZE', DEFAULT_SQLALCHEMY_DB_POOL_SIZE))


class SQLAlchemyDatastore(AbstractDatastore):

    def __init__(self, base=Base, tables=None, connection_strategy='plain',
                 session=None, **kwargs):
        super(SQLAlchemyDatastore, self).__init__(**kwargs)
        self._session = session
        self._engine = session.bind if session else None
        self._base = base
        self._tables = tables
        self._connection_strategy = connection_strategy

    @property
    def session(self):
        if self._session is None:
            if self._engine is None:
                self.setup_connection()
            session_factory = sessionmaker(bind=self._engine)
            self._session = scoped_session(session_factory)
        return self._session

    def setup_connection(self):
        assert isinstance(self.settings, SQLAlchemySettings), self.settings
        if self._engine is None:
            if self.is_sqlite():
                kwargs = {
                    'connect_args': {'check_same_thread': False},
                }
            elif self.settings.pool_size == 1:
                kwargs = {
                    'poolclass': StaticPool
                }
            else:
                kwargs = {
                    'pool_size': self.settings.pool_size,
                }
            self._engine = create_engine(
                self.settings.uri,
                strategy=self._connection_strategy,
                # echo=True,
                **kwargs
            )
            assert self._engine

    # Experiment with shared cache in sqlite :memory: database. The snag is
    # that when the database is locked by one thread and another attempts to
    # access, then the other thread immediately raises an exception because
    # the database is locked. There are no concurrent transactions, so access
    # would need to be controlled with a lock.
    #
    # def setup_connection(self):
    #     assert isinstance(self.settings, SQLAlchemySettings), self.settings
    #     if self._engine is None:
    #         args = []
    #         kwargs = {}
    #         if self.is_sqlite():
    #             kwargs['connect_args'] = {
    #                 'check_same_thread': False,
    #             }
    #
    #             if self.settings.uri == SQLITE_IN_MEMORY:
    #                 # Do some things to pass in cache=shared, so in-memory database
    #                 # has a shared (rather than private) cache, which makes it work
    #                 # with multiple threads.
    #                 PY2 = sys.version_info.major == 2
    #                 if PY2:
    #                     params = {}
    #                 else:
    #                     params = {'uri': True}
    #                     kwargs['creator'] = lambda: sqlite3.connect('file::memory:?cache=shared', **params)
    #                 args.append('sqlite://')
    #             else:
    #                 args.append(self.settings.uri)
    #
    #         elif self.settings.pool_size == 1:
    #             kwargs['poolclass'] = StaticPool
    #             args.append(self.settings.uri)
    #         else:
    #             kwargs['pool_size'] = self.settings.pool_size
    #             args.append(self.settings.uri)
    #
    #         self._engine = create_engine(
    #             strategy=self._connection_strategy,
    #             *args,
    #             **kwargs
    #         )
    #         assert self._engine

    def is_sqlite(self):
        return self.settings.uri.startswith('sqlite')

    def setup_tables(self, tables=None):
        if self._tables is not None:
            for table in self._tables:
                self.setup_table(table)

    def setup_table(self, table):
        if self._engine is None:
            raise Exception("Engine not set when required: {}".format(self))
        try:
            table.__table__.create(self._engine, checkfirst=True)
        except InternalError as e:
            if "Table '{}' already exists".format(table.__tablename__) in str(e):
                # This is a race condition from checkfirst=True. Can happen
                # if two threads call this method at the same time.
                pass
            else:
                raise

    def drop_tables(self):
        if self._tables is not None:
            for table in self._tables:
                self.drop_table(table)

    def drop_table(self, table):
        table.__table__.drop(self._engine, checkfirst=True)

    def truncate_tables(self):
        self.drop_tables()

    def close_connection(self):
        if self._session:
            self._session.close()
            self._session = None
        if self._engine:
            # Call dispose(), unless sqlite (to avoid error 'stored_events'
            # table does not exist in projections.rst doc).
            if not self.is_sqlite():
                self._engine.dispose()
            self._engine = None



from eventsourcing.tests.sequenced_item_tests.test_django_record_manager import DjangoTestCase

from eventsourcing.application.django import DjangoApplication
from eventsourcing.tests.test_system import TestSystem


class TestSystemWithDjango(DjangoTestCase, TestSystem):
    infrastructure_class = DjangoApplication

    def test_multiprocessing_multiapp_system(self):
        super(TestSystemWithDjango, self).test_multiprocessing_multiapp_system()

    def set_db_uri(self):
        # The Django settings module doesn't currently recognise DB_URI.
        pass


# Avoid running imported test case.
del (TestSystem)



from sqlalchemy import asc, bindparam, desc, select, text
from sqlalchemy.exc import DBAPIError, IntegrityError
from sqlalchemy.orm.exc import MultipleResultsFound, NoResultFound
from sqlalchemy.sql import func

from eventsourcing.exceptions import ProgrammingError
from eventsourcing.infrastructure.base import SQLRecordManager


class SQLAlchemyRecordManager(SQLRecordManager):

    _where_application_name_tmpl = (
        " WHERE application_name=:application_name AND pipeline_id=:pipeline_id"
    )

    def __init__(self, session, *args, **kwargs):
        super(SQLAlchemyRecordManager, self).__init__(*args, **kwargs)
        self.session = session

    def _prepare_insert(self, tmpl, record_class, field_names, placeholder_for_id=False):
        """
        With transaction isolation level of "read committed" this should
        generate records with a contiguous sequence of integer IDs, assumes
        an indexed ID column, the database-side SQL max function, the
        insert-select-from form, and optimistic concurrency control.
        """
        field_names = list(field_names)
        if hasattr(record_class, 'application_name') and 'application_name' not in field_names:
            field_names.append('application_name')
        if hasattr(record_class, 'pipeline_id') and 'pipeline_id' not in field_names:
            field_names.append('pipeline_id')
        if hasattr(record_class, 'causal_dependencies') and 'causal_dependencies' not in field_names:
            field_names.append('causal_dependencies')
        if self.notification_id_name:
            if placeholder_for_id:
                if self.notification_id_name not in field_names:
                    field_names.append(self.notification_id_name)

        statement = text(tmpl.format(
            tablename=self.get_record_table_name(record_class),
            columns=", ".join(field_names),
            placeholders=", ".join([":{}".format(f) for f in field_names]),
            notification_id=self.notification_id_name
        ))

        # Define bind parameters with explicit types taken from record column types.
        bindparams = []
        for col_name in field_names:
            column_type = getattr(record_class, col_name).type
            bindparams.append(bindparam(col_name, type_=column_type))

        # Redefine statement with explicitly typed bind parameters.
        statement = statement.bindparams(*bindparams)

        return statement

    def write_records(self, records, tracking_kwargs=None, orm_objs=None):
        all_params = []
        statement = None
        if records:
            # Prepare to insert event and notification records.
            statement = self.insert_values
            if self.notification_id_name:
                all_ids = set((getattr(r, self.notification_id_name) for r in records))
                if None in all_ids:
                    if len(all_ids) > 1:
                        # Either all or zero records must have IDs.
                        raise ProgrammingError("Only some records have IDs")

                    elif self.contiguous_record_ids:
                        # Do an "insert select max" from existing.
                        statement = self.insert_select_max

                    elif hasattr(self.record_class, 'application_name'):
                        # Can't allow auto-incrementing ID if table has field
                        # application_name. We need values and don't have them.
                        raise ProgrammingError("record ID not set when required")

            for record in records:
                # Params for stored item itself (e.g. event).
                params = {
                    name: getattr(record, name) for name in self.field_names
                }

                # Params for application partition (bounded context).
                if hasattr(self.record_class, 'application_name'):
                    params['application_name'] = self.application_name

                # Params for notification log.
                if self.notification_id_name:
                    notification_id = getattr(record, self.notification_id_name)
                    if notification_id == 'event-not-notifiable':
                        params[self.notification_id_name] = None
                    else:
                        params[self.notification_id_name] = notification_id

                    if hasattr(self.record_class, 'pipeline_id'):
                        if notification_id == 'event-not-notifiable':
                            params['pipeline_id'] = None
                        else:
                            params['pipeline_id'] = self.pipeline_id

                if hasattr(record, 'causal_dependencies'):
                    params['causal_dependencies'] = record.causal_dependencies

                all_params.append(params)

        try:
            nothing_to_commit = True

            # Commit custom ORM objects.
            if orm_objs:
                for orm_obj in orm_objs:
                    self.session.add(orm_obj)
                nothing_to_commit = False

            # Insert tracking record.
            if tracking_kwargs:
                self.session.execute(self.insert_tracking_record, tracking_kwargs)
                nothing_to_commit = False

            # Bulk insert event records.
            if all_params:
                self.session.execute(statement, all_params)
                nothing_to_commit = False

            if nothing_to_commit:
                return

            self.session.commit()

        except IntegrityError as e:
            self.session.rollback()
            self.raise_record_integrity_error(e)

        except DBAPIError as e:
            self.session.rollback()
            self.raise_operational_error(e)

        except:
            self.session.rollback()
            raise

        finally:
            self.session.close()


    def get_records(self, sequence_id, gt=None, gte=None, lt=None, lte=None, limit=None,
                    query_ascending=True, results_ascending=True):
        assert limit is None or limit >= 1, limit
        try:
            # Filter by sequence_id.
            filter_kwargs = {
                self.field_names.sequence_id: sequence_id
            }
            # Optionally, filter by application_name.
            if hasattr(self.record_class, 'application_name'):
                filter_kwargs['application_name'] = self.application_name
            query = self.filter_by(**filter_kwargs)

            # Filter and order by position.
            position_field = getattr(self.record_class, self.field_names.position)

            if query_ascending:
                query = query.order_by(asc(position_field))
            else:
                query = query.order_by(desc(position_field))

            if gt is not None:
                query = query.filter(position_field > gt)
            if gte is not None:
                query = query.filter(position_field >= gte)
            if lt is not None:
                query = query.filter(position_field < lt)
            if lte is not None:
                query = query.filter(position_field <= lte)

            # Limit the results.
            if limit is not None:
                query = query.limit(limit)

            # Get the result set.
            results = query.all()

        finally:
            self.session.close()

        # Reverse if necessary.
        if results_ascending != query_ascending:
            # This code path is under test, but not otherwise used ATM.
            results.reverse()

        return results

    def get_notifications(self, start=None, stop=None, *args, **kwargs):
        try:
            query = self.orm_query()
            if hasattr(self.record_class, 'application_name'):
                query = query.filter(self.record_class.application_name == self.application_name)
            if hasattr(self.record_class, 'pipeline_id'):
                query = query.filter(self.record_class.pipeline_id == self.pipeline_id)
            if self.notification_id_name:
                query = query.order_by(asc(self.notification_id_name))
                # NB '+1' because record IDs start from 1.
                notification_id_col = getattr(self.record_class, self.notification_id_name)
                if start is not None:
                    query = query.filter(notification_id_col >= start + 1)
                if stop is not None:
                    query = query.filter(notification_id_col < stop + 1)
            # Todo: Should some tables with an ID not be ordered by ID?
            # Todo: Which order do other tables have?
            return query.all()
        finally:
            self.session.close()

    def get_record(self, sequence_id, position):
        try:
            filter_args = {
                self.field_names.sequence_id: sequence_id
            }

            query = self.filter_by(**filter_args)
            if hasattr(self.record_class, 'application_name'):
                query = query.filter(
                    self.record_class.application_name == self.application_name
                )

            position_field = getattr(self.record_class, self.field_names.position)

            query = query.filter(position_field == position)
            return query.one()
        except (NoResultFound, MultipleResultsFound):
            raise IndexError(self.application_name, sequence_id, position)

    def filter_by(self, **kwargs):
        return self.orm_query().filter_by(**kwargs)

    def orm_query(self):
        return self.session.query(self.record_class)

    def get_max_record_id(self):
        try:
            notification_id_col = getattr(self.record_class, self.notification_id_name)
            query = self.session.query(func.max(notification_id_col))
            if hasattr(self.record_class, 'application_name'):
                query = query.filter(self.record_class.application_name == self.application_name)
            if hasattr(self.record_class, 'pipeline_id'):
                query = query.filter(self.record_class.pipeline_id == self.pipeline_id)

            return query.scalar()
        finally:
            self.session.close()

    def get_max_tracking_record_id(self, upstream_application_name):
        query = self.session.query(func.max(self.tracking_record_class.notification_id))
        query = query.filter(self.tracking_record_class.application_name == self.application_name)
        query = query.filter(self.tracking_record_class.upstream_application_name == upstream_application_name)
        query = query.filter(self.tracking_record_class.pipeline_id == self.pipeline_id)
        return query.scalar()

    def has_tracking_record(self, upstream_application_name, pipeline_id, notification_id):
        query = self.session.query(self.tracking_record_class)
        query = query.filter(self.tracking_record_class.application_name == self.application_name)
        query = query.filter(self.tracking_record_class.upstream_application_name == upstream_application_name)
        query = query.filter(self.tracking_record_class.pipeline_id == pipeline_id)
        query = query.filter(self.tracking_record_class.notification_id == notification_id)
        try:
            query.one()
        except (MultipleResultsFound, NoResultFound):
            return False
        else:
            return True

    def all_sequence_ids(self):
        c = self.record_class.__table__.c
        sequence_id_col = getattr(c, self.field_names.sequence_id)
        expr = select([sequence_id_col], distinct=True)

        if hasattr(self.record_class, 'application_name'):
            expr = expr.where(c.application_name == self.application_name)

        try:
            for row in self.session.query(expr):
                yield row[0]
        finally:
            self.session.close()

    def delete_record(self, record):
        """
        Permanently removes record from table.
        """
        try:
            self.session.delete(record)
            self.session.commit()
        except Exception as e:
            self.session.rollback()
            raise ProgrammingError(e)
        finally:
            self.session.close()

    def get_record_table_name(self, record_class):
        return record_class.__table__.name

    def clone(self, **kwargs):
        return super(SQLAlchemyRecordManager, self).clone(session=self.session, **kwargs)



import itertools
from copy import deepcopy
from uuid import UUID

from eventsourcing.application.process import ProcessApplication, ProcessEvent
from eventsourcing.application.system import System
from eventsourcing.contrib.paxos.composable import PaxosInstance, PaxosMessage, Resolution
from eventsourcing.domain.model.aggregate import AggregateRoot
from eventsourcing.domain.model.decorators import retry
from eventsourcing.exceptions import OperationalError, RecordConflictError, RepositoryKeyError


class PaxosAggregate(AggregateRoot):
    """
    Event-sourced Paxos participant.
    """
    paxos_variables = [
        'proposed_value',
        'proposal_id',
        'promised_id',
        'accepted_id',
        'accepted_value',
        'highest_proposal_id',
        'promises_received',
        'nacks_received',
        'highest_accepted_id',
        'leader',
        'proposals',
        'acceptors',
        'final_value',
    ]
    is_verbose = False

    def __init__(self, quorum_size, network_uid, **kwargs):
        assert isinstance(quorum_size, int)
        self.quorum_size = quorum_size
        self.network_uid = network_uid
        self.promises_received = set()
        self.nacks_received = set()
        self.leader = False
        self.proposals = {}
        self.acceptors = {}
        self.final_value = None
        super(PaxosAggregate, self).__init__(**kwargs)

    @property
    def paxos_instance(self):
        """
        Returns instance of PaxosInstance (protocol implementation).
        """
        # Construct instance with the constant attributes.
        instance = PaxosInstance(self.network_uid, self.quorum_size)

        # Set the variable attributes from the aggregate.
        for name in self.paxos_variables:
            value = getattr(self, name, None)
            if value is not None:
                if isinstance(value, (set, list, dict, tuple)):
                    value = deepcopy(value)
                setattr(instance, name, value)

        # Return the instance.
        return instance

    class Event(AggregateRoot.Event):
        """
        Base event class for PaxosAggregate.
        """

    class Started(Event, AggregateRoot.Created):
        """
        Published when a PaxosAggregate is started.
        """
        __notifiable__ = False

    class AttributesChanged(Event):
        """
        Published when attributes of paxos_instance are changed.
        """
        __notifiable__ = False

        def __init__(self, changes=None, **kwargs):
            super(PaxosAggregate.AttributesChanged, self).__init__(
                changes=changes, **kwargs
            )

        @property
        def changes(self):
            return self.__dict__['changes']

        def mutate(self, obj):
            for name, value in self.changes.items():
                setattr(obj, name, value)

    class MessageAnnounced(Event):
        """
        Published when a Paxos message is announced.
        """

        @property
        def msg(self):
            return self.__dict__['msg']

    @classmethod
    def start(cls, originator_id, quorum_size, network_uid):
        """
        Factory method that returns a new Paxos aggregate.
        """
        assert isinstance(quorum_size, int), "Not an integer: {}".format(quorum_size)
        return cls.__create__(
            event_class=cls.Started,
            originator_id=originator_id,
            quorum_size=quorum_size,
            network_uid=network_uid
        )

    def propose_value(self, value, assume_leader=False):
        """
        Proposes a value to the network.
        """
        if value is None:
            raise ValueError("Not allowed to propose value None")
        paxos = self.paxos_instance
        paxos.leader = assume_leader
        msg = paxos.propose_value(value)
        if msg is None:
            msg = paxos.prepare()
        self.setattrs_from_paxos(paxos)
        self.announce(msg)
        return msg

    def receive_message(self, msg):
        """
        Responds to messages from other participants.
        """
        if isinstance(msg, Resolution):
            return
        paxos = self.paxos_instance
        while msg:
            if isinstance(msg, Resolution):
                self.print_if_verbose("{} resolved value {}".format(self.network_uid, msg.value))
                break
            else:
                self.print_if_verbose("{} <- {} <- {}".format(self.network_uid, msg.__class__.__name__, msg.from_uid))
                msg = paxos.receive(msg)
                # Todo: Make it optional not to announce resolution (without which it's hard to see final value).
                do_announce_resolution = True
                if msg and (do_announce_resolution or not isinstance(msg, Resolution)):
                    self.announce(msg)

        self.setattrs_from_paxos(paxos)

    def announce(self, msg):
        """
        Announces a Paxos message.
        """
        self.print_if_verbose("{} -> {}".format(self.network_uid, msg.__class__.__name__))
        self.__trigger_event__(
            event_class=self.MessageAnnounced,
            msg=msg,
        )

    def setattrs_from_paxos(self, paxos):
        """
        Registers changes of attribute value on Paxos instance.
        """
        changes = {}
        for name in self.paxos_variables:
            paxos_value = getattr(paxos, name)
            if paxos_value != getattr(self, name, None):
                self.print_if_verbose("{} {}: {}".format(self.network_uid, name, paxos_value))
                changes[name] = paxos_value
                setattr(self, name, paxos_value)
        if changes:
            self.__trigger_event__(
                event_class=self.AttributesChanged,
                changes=changes
            )

    def print_if_verbose(self, param):
        if self.is_verbose:
            print(param)

    def __str__(self):
        return ("PaxosAggregate("
                "network_uid='{network_uid}', "
                "proposal_id='{proposal_id}', "
                "promised_id='{promised_id}', "
                "promises_received='{promises_received}'"
                ")").format(**self.__dict__)


class PaxosProcess(ProcessApplication):
    persist_event_type = PaxosAggregate.Event
    quorum_size = None
    notification_log_section_size = 5
    use_cache = True

    # Todo: Reintroduce this, if it can be made to work with Popo infrastructure.
    # set_notification_ids = True

    @retry((RecordConflictError, OperationalError), max_attempts=10, wait=0)
    def propose_value(self, key, value, assume_leader=False):
        """
        Starts new Paxos aggregate and proposes a value for a key.

        Decorated with retry in case of notification log conflict
        or operational error.
        """
        assert isinstance(key, UUID)
        paxos_aggregate = PaxosAggregate.start(
            originator_id=key,
            quorum_size=self.quorum_size,
            network_uid=self.name
        )
        msg = paxos_aggregate.propose_value(value, assume_leader=assume_leader)
        while msg:
            msg = paxos_aggregate.receive_message(msg)
        new_events = paxos_aggregate.__batch_pending_events__()
        self.record_process_event(ProcessEvent(new_events))
        self.repository.take_snapshot(paxos_aggregate.id)
        self.publish_prompt()
        return paxos_aggregate  # in case it's new

    def policy(self, repository, event):
        if isinstance(event, PaxosAggregate.MessageAnnounced):
            msg = event.msg
            assert isinstance(msg, PaxosMessage)
            # Get or create aggregate.
            try:
                paxos = repository[event.originator_id]
            except RepositoryKeyError:
                paxos = PaxosAggregate.start(
                    originator_id=event.originator_id,
                    quorum_size=self.quorum_size,
                    network_uid=self.name
                )
                # Needs to go in the cache now, otherwise we get
                # "Duplicate" errors (for some unknown reason).
                self.repository._cache[paxos.id] = paxos
            assert isinstance(paxos, PaxosAggregate), type(paxos)
            # Absolutely make sure the participant aggregates aren't getting confused.
            assert paxos.network_uid == self.name, (
                "Wrong paxos aggregate: required network_uid {}, got {}".format(
                    self.name, paxos.network_uid
                )
            )
            # Only receive messages until resolution is
            # obtained. Followers will process our previous
            # announcements and resolve to the same final value
            # before processing anything we could announce after.
            if paxos.final_value is None:
                paxos.receive_message(msg)

            return paxos


class PaxosSystem(System):
    def __init__(self, num_participants=3, **kwargs):
        self.num_participants = num_participants
        self.quorum_size = (num_participants + 2) // 2
        classes = [
            type(
                'PaxosProcess{}'.format(i),
                (PaxosProcess,),
                {'quorum_size': self.quorum_size}
            ) for i in range(num_participants)
        ]
        if num_participants > 1:
            pipelines = [[c[0], c[1], c[0]] for c in itertools.combinations(classes, 2)]
        else:
            pipelines = [classes]
        super(PaxosSystem, self).__init__(*pipelines, **kwargs)



from unittest import TestCase
from uuid import uuid4

from eventsourcing.domain.model.events import DomainEvent
from eventsourcing.infrastructure.eventstore import EventStore
from eventsourcing.infrastructure.sqlalchemy.datastore import SQLAlchemyDatastore, SQLAlchemySettings
from eventsourcing.infrastructure.sqlalchemy.factory import construct_sqlalchemy_eventstore
from eventsourcing.infrastructure.sqlalchemy.records import StoredEventRecord


class TestFactory(TestCase):
    def test_construct_sqlalchemy_eventstore(self):
        # Construct the infrastructure.
        datastore = SQLAlchemyDatastore(settings=SQLAlchemySettings())
        datastore.setup_connection()

        event_store = construct_sqlalchemy_eventstore(
            session=datastore.session,
            contiguous_record_ids=True,
            application_name=uuid4().hex,
        )
        datastore.setup_table(event_store.record_manager.record_class)

        self.assertIsInstance(event_store, EventStore)

        # Store an event.
        aggregate_id = uuid4()
        aggregate_version = 0
        domain_event = DomainEvent(
            a=1,
            originator_id=aggregate_id,
            originator_version=aggregate_version,
        )
        event_store.store(domain_event)

        # Get the domain events.
        events = event_store.get_domain_events(originator_id=aggregate_id)
        self.assertEqual(len(events), 1)
        self.assertEqual(events[0], domain_event)

        # Test the while clause of all_sequence_ids() is filtering on application_name.
        self.assertEqual(event_store.record_manager.record_class, StoredEventRecord)
        sequence_ids = event_store.record_manager.list_sequence_ids()
        self.assertEqual([aggregate_id], sequence_ids)

        # - check the aggregate ID isn't listed by another application
        event_store = construct_sqlalchemy_eventstore(
            session=datastore.session,
            contiguous_record_ids=True,
            application_name=uuid4().hex,
        )
        sequence_ids = event_store.record_manager.list_sequence_ids()
        self.assertEqual([], sequence_ids)



import datetime
import os
import unittest
from time import sleep
from uuid import uuid4

from eventsourcing.application.multiprocess import MultiprocessRunner
from eventsourcing.application.sqlalchemy import SQLAlchemyApplication
from eventsourcing.application.system import MultiThreadedRunner
from eventsourcing.contrib.paxos.application import PaxosSystem, PaxosProcess
from eventsourcing.domain.model.decorators import retry
from eventsourcing.domain.model.events import assert_event_handlers_empty, clear_event_handlers
from eventsourcing.tests.base import notquick
from eventsourcing.tests.test_system_fixtures import set_db_uri


class TestPaxosSystem(unittest.TestCase):

    # Use the same system object in all tests.
    system = PaxosSystem(setup_tables=True)

    # Use SQLAlchemy infrastructure (override in subclasses).
    infrastructure_class = SQLAlchemyApplication

    def test_single_threaded(self):

        key1, key2, key3 = uuid4(), uuid4(), uuid4()
        value1, value2, value3 = 11111, 22222, 33333

        with self.system.bind(self.infrastructure_class) as runner:
            paxosprocess0 = runner.paxosprocess0
            paxosprocess1 = runner.paxosprocess1
            paxosprocess2 = runner.paxosprocess2

            started1 = datetime.datetime.now()
            assert isinstance(paxosprocess0, PaxosProcess)
            paxosprocess0.propose_value(key1, value1)
            ended1 = (datetime.datetime.now() - started1).total_seconds()

            # Check each process has expected final value.
            self.assert_final_value(paxosprocess0, key1, value1)
            self.assert_final_value(paxosprocess1, key1, value1)
            self.assert_final_value(paxosprocess2, key1, value1)
            print("Resolved paxos 1 with single thread in %ss" % ended1)

            started2 = datetime.datetime.now()
            paxosprocess1.propose_value(key2, value2)
            ended2 = (datetime.datetime.now() - started2).total_seconds()

            # Check each process has a resolution.
            self.assert_final_value(paxosprocess0, key2, value2)
            self.assert_final_value(paxosprocess1, key2, value2)
            self.assert_final_value(paxosprocess2, key2, value2)
            print("Resolved paxos 2 with single thread in %ss" % ended2)

            started3 = datetime.datetime.now()
            paxosprocess2.propose_value(key3, value3)
            ended3 = (datetime.datetime.now() - started3).total_seconds()

            # Check each process has a resolution.
            self.assert_final_value(paxosprocess0, key3, value3)
            self.assert_final_value(paxosprocess1, key3, value3)
            self.assert_final_value(paxosprocess2, key3, value3)
            print("Resolved paxos 3 with single thread in %ss" % ended3)

    def test_multi_threaded(self):

        if 'TRAVIS_PYTHON_VERSION' in os.environ:
            if self.infrastructure_class is SQLAlchemyApplication:
                self.skipTest("There's an intermittent problem with the multi-threaded"
                              "runner with SQLAlchemy and Python 3.7 on Travis. Fix me :).")
        set_db_uri()

        key1, key2, key3 = uuid4(), uuid4(), uuid4()
        value1, value2, value3 = 11111, 22222, 33333

        runner = MultiThreadedRunner(
            system=self.system,
            infrastructure_class=self.infrastructure_class
        )

        with runner:
            paxosprocess0 = runner.processes['paxosprocess0']
            paxosprocess1 = runner.processes['paxosprocess1']
            paxosprocess2 = runner.processes['paxosprocess2']

            started1 = datetime.datetime.now()
            assert isinstance(paxosprocess0, PaxosProcess)
            paxosprocess0.propose_value(key1, value1)
            ended1 = (datetime.datetime.now() - started1).total_seconds()

            # Check each process has expected final value.
            self.assert_final_value(paxosprocess0, key1, value1)
            self.assert_final_value(paxosprocess1, key1, value1)
            self.assert_final_value(paxosprocess2, key1, value1)
            print("Resolved paxos 1 with multi threads in %ss" % ended1)

            started2 = datetime.datetime.now()
            paxosprocess1.propose_value(key2, value2)
            ended2 = (datetime.datetime.now() - started2).total_seconds()

            # Check each process has a resolution.
            self.assert_final_value(paxosprocess0, key2, value2)
            self.assert_final_value(paxosprocess1, key2, value2)
            self.assert_final_value(paxosprocess2, key2, value2)
            print("Resolved paxos 2 with multi threads in %ss" % ended2)

            started3 = datetime.datetime.now()
            paxosprocess2.propose_value(key3, value3)
            ended3 = (datetime.datetime.now() - started3).total_seconds()

            # Check each process has a resolution.
            self.assert_final_value(paxosprocess0, key3, value3)
            self.assert_final_value(paxosprocess1, key3, value3)
            self.assert_final_value(paxosprocess2, key3, value3)
            print("Resolved paxos 3 with multi threads in %ss" % ended3)

    def test_multiprocessing(self):
        set_db_uri()

        key1, key2, key3 = uuid4(), uuid4(), uuid4()
        value1, value2, value3 = 11111, 22222, 33333

        self.close_connections_before_forking()

        pipeline_ids = [1, 2, 3]
        runner = MultiprocessRunner(
            system=self.system,
            pipeline_ids=pipeline_ids,
            infrastructure_class=self.infrastructure_class
        )

        # Start running operating system processes.
        with runner:
            # Get local application object.
            paxosprocess0 = runner.paxosprocess0
            assert isinstance(paxosprocess0, PaxosProcess)

            # Start proposing values on the different system pipelines.
            paxosprocess0.change_pipeline(1)
            started1 = datetime.datetime.now()
            paxosprocess0.propose_value(key1, value1)

            paxosprocess0.change_pipeline(2)
            started2 = datetime.datetime.now()
            paxosprocess0.propose_value(key2, value2)

            paxosprocess0.change_pipeline(3)
            started3 = datetime.datetime.now()
            paxosprocess0.propose_value(key3, value3)

            # Check all the process applications have expected final values.
            paxosprocess1 = runner.paxosprocess1
            paxosprocess2 = runner.paxosprocess1
            assert isinstance(paxosprocess1, PaxosProcess)
            paxosprocess0.repository.use_cache = False
            paxosprocess1.repository.use_cache = False
            paxosprocess2.repository.use_cache = False

            self.assert_final_value(paxosprocess0, key1, value1)
            self.assert_final_value(paxosprocess1, key1, value1)
            self.assert_final_value(paxosprocess2, key1, value1)
            duration1 = (datetime.datetime.now() - started1).total_seconds()
            print("Resolved paxos 1 with multiprocessing in %ss" % duration1)

            self.assert_final_value(paxosprocess0, key2, value2)
            self.assert_final_value(paxosprocess1, key2, value2)
            self.assert_final_value(paxosprocess2, key2, value2)
            duration2 = (datetime.datetime.now() - started2).total_seconds()
            print("Resolved paxos 2 with multiprocessing in %ss" % duration2)

            self.assert_final_value(paxosprocess0, key3, value3)
            self.assert_final_value(paxosprocess1, key3, value3)
            self.assert_final_value(paxosprocess2, key3, value3)
            duration3 = (datetime.datetime.now() - started3).total_seconds()
            print("Resolved paxos 3 with multiprocessing in %ss" % duration3)

    @notquick
    def test_multiprocessing_performance(self):

        set_db_uri()

        num_pipelines = 2
        pipeline_ids = range(num_pipelines)

        runner = MultiprocessRunner(
            system=self.system,
            pipeline_ids=pipeline_ids,
            infrastructure_class=self.infrastructure_class,
            setup_tables = True
        )

        num_proposals = 50

        self.close_connections_before_forking()

        with runner:
            sleep(1)

            # Construct an application instance in this process.
            paxos_process = runner.paxosprocess0
            assert isinstance(paxos_process, PaxosProcess)

            # Don't use the cache, so as to keep checking actual database.
            paxos_process.repository.use_cache = False

            # Start timing (just for fun).
            started = datetime.datetime.now()

            # Propose values.
            proposals = list(((uuid4(), i) for i in range(num_proposals)))
            for key, value in proposals:
                paxos_process.change_pipeline((value % len(pipeline_ids)))
                print("Proposing key {} value {}".format(key, value))
                paxos_process.propose_value(key, str(value))
                sleep(0.0)

            # Check final values.
            for key, value in proposals:
                print("Asserting final value for key {} value {}".format(key, value))
                self.assert_final_value(paxos_process, key, str(value))

            # Print timing information (just for fun).
            duration = (datetime.datetime.now() - started).total_seconds()
            print(
                "Resolved {} paxoses with multiprocessing in {:.4f}s ({:.4f}s each)".format(
                    num_proposals, duration, duration / num_proposals
                )
            )

    @retry((KeyError, AssertionError), max_attempts=100, wait=0.2, stall=0)
    def assert_final_value(self, process, id, value):
        self.assertEqual(process.repository[id].final_value, value)

    def close_connections_before_forking(self):
        """Implemented by the DjangoTestCase class."""
        pass

    def setUp(self) -> None:
        assert_event_handlers_empty()

    def tearDown(self):
        try:
            del (os.environ['DB_URI'])
        except KeyError:
            pass
        try:
            assert_event_handlers_empty()
        finally:
            clear_event_handlers()



import time
from collections import OrderedDict, defaultdict
from threading import Lock

from eventsourcing.application.notificationlog import NotificationLogReader
from eventsourcing.application.simple import SimpleApplication
from eventsourcing.application.snapshotting import SnapshottingApplication
from eventsourcing.domain.model.events import publish, subscribe, unsubscribe
from eventsourcing.exceptions import CausalDependencyFailed, PromptFailed
from eventsourcing.infrastructure.base import ACIDRecordManager
from eventsourcing.infrastructure.eventsourcedrepository import EventSourcedRepository
from eventsourcing.utils.transcoding import json_dumps, json_loads


class ProcessEvent(object):
    def __init__(self, new_events, tracking_kwargs=None, causal_dependencies=None, orm_objs=None):
        self.new_events = new_events
        self.tracking_kwargs = tracking_kwargs
        self.causal_dependencies = causal_dependencies
        self.orm_objs = orm_objs


class ProcessApplication(SimpleApplication):
    set_notification_ids = False
    use_causal_dependencies = False
    notification_log_reader_class = NotificationLogReader

    def __init__(self, name=None, policy=None, setup_table=False,
                 use_direct_query_if_available=False,
                 notification_log_reader_class=None,
                 **kwargs):
        self.policy_func = policy
        self.readers = OrderedDict()
        self.is_reader_position_ok = defaultdict(bool)
        self._notification_generators = {}
        self._policy_lock = Lock()
        self.clock_event = None
        self.tick_interval = None
        self.use_direct_query_if_available = use_direct_query_if_available
        self.notification_log_reader_class = notification_log_reader_class or \
                                             type(self).notification_log_reader_class

        super(ProcessApplication, self).__init__(name=name, setup_table=setup_table, **kwargs)

        if self.event_store:
            self.notification_topic_key = self.event_store.record_manager.field_names.topic
            self.notification_state_key = self.event_store.record_manager.field_names.state

        # Publish prompts for any domain events that we persist.
        if self.persistence_policy:
            subscribe(
                predicate=self.persistence_policy.is_event,
                handler=self.publish_prompt,
            )

    def close(self):
        if self.persistence_policy:
            unsubscribe(
                predicate=self.persistence_policy.is_event,
                handler=self.publish_prompt,
            )
        super(ProcessApplication, self).close()

    def publish_prompt(self, event=None):
        """
        Publishes prompt for a given event.

        Used to prompt downstream process application when an event
        is published by this application's model, which can happen
        when application command methods, rather than the process policy,
        are called.

        Wraps exceptions with PromptFailed, to avoid application policy exceptions being
        seen directly in other applications when running synchronously in single thread.
        """

        prompt = Prompt(self.name, self.pipeline_id)
        try:
            publish(prompt)
        except PromptFailed:
            raise
        except Exception as e:
            raise PromptFailed("{}: {}".format(type(e), str(e)))

    def follow(self, upstream_application_name, notification_log):
        # Create a reader.
        reader = self.notification_log_reader_class(
            notification_log,
            use_direct_query_if_available=self.use_direct_query_if_available
        )
        self.readers[upstream_application_name] = reader

    def run(self, prompt=None, advance_by=None):

        if prompt:
            assert isinstance(prompt, Prompt)
            upstream_names = [prompt.process_name]
        else:
            upstream_names = self.readers.keys()

        notification_count = 0
        for upstream_name in upstream_names:

            if not self.is_reader_position_ok[upstream_name]:
                self.del_notification_generator(upstream_name)
                self.set_reader_position_from_tracking_records(upstream_name)
                self.is_reader_position_ok[upstream_name] = True

            while True:
                with self._policy_lock:
                    # Get notification generator.
                    generator = self.get_notification_generator(upstream_name, advance_by)
                    try:
                        notification = next(generator)
                    except StopIteration:
                        self.del_notification_generator(upstream_name)
                        break

                    notification_count += 1

                    # Get domain event from notification.
                    event = self.get_event_from_notification(notification)

                    # Decode causal dependencies of the domain event.
                    causal_dependencies = notification.get('causal_dependencies') or '[]'
                    causal_dependencies = json_loads(causal_dependencies) or []

                    # Check causal dependencies are satisfied.
                    for causal_dependency in causal_dependencies:
                        pipeline_id = causal_dependency['pipeline_id']
                        notification_id = causal_dependency['notification_id']

                        _manager = self.event_store.record_manager
                        has_tracking_record = _manager.has_tracking_record(
                            upstream_application_name=upstream_name,
                            pipeline_id=pipeline_id,
                            notification_id=notification_id
                        )
                        if not has_tracking_record:
                            # Invalidate reader position.
                            self.is_reader_position_ok[upstream_name] = False

                            # Raise exception.
                            raise CausalDependencyFailed({
                                'application_name': self.name,
                                'upstream_name': upstream_name,
                                'pipeline_id': pipeline_id,
                                'notification_id': notification_id
                            })

                    # Wait on the clock event, if there is one.
                    if self.clock_event is not None:
                        self.clock_event.wait()

                    # print("Processing upstream event: ", event)
                    new_events = self.process_upstream_event(event, notification['id'], upstream_name)

                self.take_snapshots(new_events)

                # Publish a prompt if there are new notifications.
                # Todo: Optionally send events as prompts, saves pulling event if it arrives in order.
                if any([event.__notifiable__ for event in new_events]):
                    self.publish_prompt()

        return notification_count

    def process_upstream_event(self, event, notification_id, upstream_name):
        if self.tick_interval is not None:
            cycle_started = time.process_time()
        else:
            cycle_started = None
        # Call policy with the upstream event.
        all_aggregates, causal_dependencies, orm_objs = self.call_policy(event)
        # Collect pending events.
        new_events = self.collect_pending_events(all_aggregates)
        # Record process event.
        try:
            tracking_kwargs = self.construct_tracking_kwargs(
                notification_id, upstream_name
            )
            process_event = ProcessEvent(
                new_events=new_events,
                tracking_kwargs=tracking_kwargs,
                causal_dependencies=causal_dependencies,
                orm_objs=orm_objs
            )
            self.record_process_event(process_event)

            # Todo: Maybe write one tracking record at the end of a run, if
            # necessary, or only during a period of time when nothing happens?
        except Exception as exc:
            # Need to invalidate reader position, so it is refreshed.
            self.is_reader_position_ok[upstream_name] = False

            # Need to purge from the cache relevant entities that
            # have evolved their state past what has been recorded,
            # otherwise strange errors (about version mismatches, or
            # when identifying causal dependencies) can arise.
            if self.repository._cache:
                originator_ids = set([event.originator_id for event in new_events])
                for originator_id in originator_ids:
                    try:
                        del self.repository._cache[originator_id]
                    except KeyError:
                        pass
            raise exc
        else:
            if self.tick_interval is not None:
                # Todo: Change this to use the full cycle time (improve getting notifications first).
                cycle_ended = time.process_time()
                cycle_time = cycle_ended - cycle_started
                cycle_perc = 100 * (cycle_time) / self.tick_interval
                if cycle_perc > 100:
                    msg = f"Warning: {self.name} cycle exceeded tick interval by: {cycle_perc - 100:.2f}%"
                    print(msg)
        return new_events

    def get_event_from_notification(self, notification):
        return self.event_store.mapper.event_from_topic_and_state(
            topic=notification[self.notification_topic_key],
            state=notification[self.notification_state_key]
        )

    def get_notification_generator(self, upstream_name, advance_by):
        # Dict avoids re-entrant calls to run() starting their own generator,
        # so that notifications are only received once. Was needed in
        # single-threaded runner before it was changed to use iteration not
        # recursion. Hence, probably no longer needed - use reader directly.
        try:
            generator = self._notification_generators[upstream_name]
        except KeyError:
            # Todo: Rename as 'iterator'? We use an iterator, doesn't matter whether or not it is a generator.
            generator = self.read_reader(upstream_name, advance_by)
            self._notification_generators[upstream_name] = generator
        return generator

    def read_reader(self, upstream_name, advance_by=None):
        return self.readers[upstream_name].read(advance_by=advance_by)

    def del_notification_generator(self, upstream_name):
        try:
            del self._notification_generators[upstream_name]
        except KeyError:
            pass

    def take_snapshots(self, new_events):
        pass

    def set_reader_position_from_tracking_records(self, upstream_name):
        max_record_id = self.event_store.record_manager.get_max_tracking_record_id(upstream_name)
        reader = self.readers[upstream_name]
        reader.seek(max_record_id or 0)

    def call_policy(self, event):
        # Get the application policy.
        policy = self.policy_func or self.policy

        # Wrap the actual repository, so we can collect aggregates.
        repository = WrappedRepository(self.repository)

        # Actually call the policy.
        new_aggregates = policy(repository, event)

        # Collect all aggregates.
        repo_aggregates = list(repository.retrieved_aggregates.values())
        all_aggregates = repo_aggregates[:]
        if new_aggregates is not None:
            if not isinstance(new_aggregates, (list, tuple)):
                new_aggregates = [new_aggregates]
            if self.repository._use_cache:
                for new_aggregate in new_aggregates:
                    self.repository._cache[new_aggregate.id] = new_aggregate
            all_aggregates += new_aggregates

        # Identify causal dependencies.
        causal_dependencies = []
        if self.use_causal_dependencies:
            highest = defaultdict(int)
            for entity_id, entity_version in repository.causal_dependencies:
                pipeline_id, notification_id = self.event_store.record_manager.get_pipeline_and_notification_id(
                    entity_id, entity_version
                )
                if pipeline_id is not None and pipeline_id != self.pipeline_id:
                    highest[pipeline_id] = max(notification_id, highest[pipeline_id])

            causal_dependencies = []
            for pipeline_id, notification_id in highest.items():
                causal_dependencies.append({
                    'pipeline_id': pipeline_id,
                    'notification_id': notification_id
                })
        # Todo: Optionally reference causal dependencies in current pipeline.
        # Todo: Support processing notification from a single pipeline in parallel, according to dependencies.
        return all_aggregates, causal_dependencies, repository.pending_orm_objs

    @staticmethod
    def policy(repository, event):
        """Empty method, can be overridden in subclasses to implement concrete policy."""

    def collect_pending_events(self, aggregates):
        pending_events = []
        num_changed_aggregates = 0
        # This doesn't necessarily obtain events in causal order...
        for aggregate in aggregates:
            batch = aggregate.__batch_pending_events__()
            if len(batch):
                num_changed_aggregates += 1
            pending_events += batch

        # ...so sort pending events across all aggregates.
        if num_changed_aggregates > 1:
            # Sort the events by timestamp.
            #  - this method is intended to establish the correct
            #    causal ordering of all these new events across all aggregates. It
            #    should work if all events are timestamped, all their timestamps
            #    are from the same clock, and none have the same value. If this
            #    doesn't work properly, for some reason, it would be possible when
            #    several aggregates publish events that depend on each other that
            #    concatenating pending events taken from each in turn will be incorrect
            #    and could potentially cause processing errors in a downstream process
            #    application that depends on the correct causal ordering of events. In
            #    the worst case, the events will still be placed correctly in the
            #    aggregate sequence, but if timestamps are skewed and so do not correctly
            #    order the events, the events may be out of order in their notification log.
            #    It is expected in normal usage that these events are created in the same
            #    operating system thread, with timestamps from the same operating system clock,
            #    and so the timestamps will provide the correct order. However, if somehow
            #    different events are timestamped from different clocks, then problems may occur
            #    if those clocks give timestamps that skew the correct causal order.
            pending_events.sort(key=lambda x: x.timestamp)

        return pending_events

    def construct_tracking_kwargs(self, notification_id, upstream_application_name):
        return {
            'application_name': self.name,
            'upstream_application_name': upstream_application_name,
            'pipeline_id': self.pipeline_id,
            'notification_id': notification_id,
        }

    def record_process_event(self, process_event: ProcessEvent):
        # Construct event records.
        event_records = self.construct_event_records(process_event.new_events,
                                                     process_event.causal_dependencies)

        # Write event records with tracking record.
        record_manager = self.event_store.record_manager
        assert isinstance(record_manager, ACIDRecordManager)
        record_manager.write_records(records=event_records,
                                     tracking_kwargs=process_event.tracking_kwargs,
                                     orm_objs=process_event.orm_objs)

    def construct_event_records(self, pending_events, causal_dependencies=None):
        # Convert to event records.
        sequenced_items = self.event_store.item_from_event(pending_events)
        event_records = self.event_store.record_manager.to_records(sequenced_items)

        # Set notification log IDs, and causal dependencies.
        if len(event_records):
            # Todo: Maybe keep track of what this probably is, to avoid query. Like log reader, invalidate on error.
            if self.set_notification_ids:
                current_max = self.event_store.record_manager.get_max_record_id() or 0
                for domain_event, event_record in zip(pending_events, event_records):
                    if type(domain_event).__notifiable__:
                        current_max += 1
                        event_record.id = current_max
                    else:
                        event_record.id = 'event-not-notifiable'

            if self.use_causal_dependencies:
                assert hasattr(self.event_store.record_manager.record_class, 'causal_dependencies')
                causal_dependencies = json_dumps(causal_dependencies)
                # Only need first event to carry the dependencies.
                event_records[0].causal_dependencies = causal_dependencies

        return event_records

    def setup_table(self):
        super(ProcessApplication, self).setup_table()
        if self.datastore is not None:
            self.datastore.setup_table(
                self.event_store.record_manager.tracking_record_class
            )

    def drop_table(self):
        super(ProcessApplication, self).drop_table()
        if self.datastore is not None:
            self.datastore.drop_table(
                self.event_store.record_manager.tracking_record_class
            )


class WrappedRepository(object):
    """
    Used to wrap an event sourced repository for use in process application
    policy so that use of, and changes to, domain model aggregates can be
    automatically detected and recorded.

    Implements a "dictionary like" interface, so that aggregates can be
    accessed by ID.
    """
    def __init__(self, repository: EventSourcedRepository):
        self.retrieved_aggregates = {}
        self.repository = repository
        self.causal_dependencies = []
        self.pending_orm_objs = []

    def __getitem__(self, entity_id):
        try:
            return self.retrieved_aggregates[entity_id]
        except KeyError:
            entity = self.repository.__getitem__(entity_id)
            self.retrieved_aggregates[entity_id] = entity
            self.causal_dependencies.append((entity.id, entity.__version__))
            return entity

    def __contains__(self, entity_id):
        return self.repository.__contains__(entity_id)

    def save_orm_obj(self, orm_obj):
        """
        Includes orm_obj in "process event", so that projections into
        custom ORM objects is as reliable with respect to sudden restarts
        as "normal" domain event processing in a process application.
        """
        self.pending_orm_objs.append(orm_obj)


class Prompt(object):
    def __init__(self, process_name, pipeline_id):
        self.process_name = process_name
        self.pipeline_id = pipeline_id

    def __eq__(self, other):
        return (
            other
            and isinstance(other, type(self))
            and self.process_name == other.process_name
            and self.pipeline_id == other.pipeline_id
        )

    def __repr__(self):
        return "{}({}={}, {}={})".format(
            type(self).__name__,
            'process_name', self.process_name,
            'pipeline_id', self.pipeline_id
        )


class ProcessApplicationWithSnapshotting(SnapshottingApplication, ProcessApplication):
    def take_snapshots(self, new_events):
        for event in new_events:
            if self.snapshotting_policy.condition(event):
                self.snapshotting_policy.take_snapshot(event)



from eventsourcing.exceptions import RecordConflictError
from eventsourcing.infrastructure.base import ACIDRecordManager
from readerwriterlock import rwlock


class PopoNotification(object):
    def __init__(self, notification_id, originator_id, originator_version, topic, state):
        self.notification_id = notification_id
        self.originator_id = originator_id
        self.originator_version = originator_version
        self.topic = topic
        self.state = state


class PopoRecordManager(ACIDRecordManager):
    def __init__(self, *args, **kwargs):
        super(PopoRecordManager, self).__init__(*args, **kwargs)
        self._all_sequence_records = {}
        self._all_sequence_max = {}
        self._all_tracking_records = {}
        self._all_tracking_max = {}
        self._all_notification_records = {}
        self._all_notification_max = {}
        self._rw_lock = rwlock.RWLockFair()

    def all_sequence_ids(self):
        with self._rw_lock.gen_rlock():
            try:
                return self._all_sequence_records[self.application_name].keys()
            except KeyError:
                return []

    def delete_record(self, record):
        with self._rw_lock.gen_wlock():
            sequence_records = self._get_sequence_records(record.sequence_id)
            try:
                position = getattr(record, self.field_names.position)
                del(sequence_records[position])
            except KeyError:
                pass

    def get_max_record_id(self):
        with self._rw_lock.gen_rlock():
            return self._get_max_record_id()

    def _get_max_record_id(self):
        try:
            max_notification_id = self._all_notification_max[self.application_name]
        except KeyError:
            pass
        else:
            return max_notification_id

    def _get_notification_records(self):
        try:
            notification_records = self._all_notification_records[self.application_name]
        except KeyError:
            notification_records = {}
        return notification_records

    def get_notifications(self, start=None, stop=None, *args, **kwargs):
        notifications = []
        with self._rw_lock.gen_rlock():
            notification_records = self._get_notification_records()
            i = start + 1
            while True:
                if stop is not None and i >= stop + 1:
                    break
                try:
                    notification_record = notification_records[i]
                    notification = PopoNotification(
                        notification_id=notification_record['notification_id'],
                        originator_id=notification_record['sequenced_item'].originator_id,
                        originator_version=notification_record['sequenced_item'].originator_version,
                        topic=notification_record['sequenced_item'].topic,
                        state=notification_record['sequenced_item'].state,

                    )
                    notifications.append(notification)
                except KeyError:
                    break
                else:
                    i += 1

        return notifications

    def get_max_tracking_record_id(self, upstream_application_name):
        with self._rw_lock.gen_rlock():
            try:
                tracking_records = self._all_tracking_records[self.application_name][upstream_application_name]
            except KeyError:
                pass
            else:
                if tracking_records:
                    return max(tracking_records)

    def get_record(self, sequence_id, position):
        with self._rw_lock.gen_rlock():
            try:
                return self._get_sequence_records(sequence_id)[position]
            except KeyError:
                raise IndexError(self.application_name, sequence_id, position)

    def get_records(self, sequence_id, gt=None, gte=None, lt=None, lte=None, limit=None,
                    query_ascending=True, results_ascending=True):

        start = None
        if gt is not None:
            start = gt + 1
        if gte is not None:
            if start is None:
                start = gte
            else:
                start = max(start, gte)

        end = None
        if lt is not None:
            end = lt
        if lte is not None:
            if end is None:
                end = lte + 1
            else:
                end = min(end, lte + 1)

        selected_records = []
        with self._rw_lock.gen_rlock():
            all_sequence_records = self._get_sequence_records(sequence_id)
            if not len(all_sequence_records):
                return []

            if end is None:
                max_position = self._all_sequence_max[self.application_name][sequence_id]
                end = max_position + 1
            if start is None:
                start = min(all_sequence_records.keys())

            for position in range(start, end):
                try:
                    record = all_sequence_records[position]
                except KeyError:
                    pass
                else:
                    selected_records.append(record)

            if not query_ascending:
                selected_records = reversed(selected_records)

            if limit is not None:
                selected_records = list(selected_records)[:limit]

            if query_ascending != results_ascending:
                selected_records = reversed(selected_records)

        return selected_records

    def _get_sequence_records(self, sequence_id):
        try:
            return self._all_sequence_records[self.application_name][sequence_id]
        except KeyError:
            return {}

    def has_tracking_record(self, upstream_application_name, pipeline_id, notification_id):
        raise NotImplementedError()

    def record_sequenced_items(self, sequenced_item_or_items):
        records = self.to_records(sequenced_item_or_items)
        self.write_records(records=records)

    def write_records(self, records, tracking_kwargs=None, orm_objs=None):
        with self._rw_lock.gen_wlock():
            # Write event and notification records.
            self._insert_records(records)

            if tracking_kwargs:
                # Write a tracking record.
                upstream_application_name = tracking_kwargs['upstream_application_name']
                application_name = tracking_kwargs['application_name']
                notification_id = tracking_kwargs['notification_id']
                assert application_name == self.application_name, (application_name, self.application_name)
                try:
                    app_tracking_records = self._all_tracking_records[application_name]
                except KeyError:
                    app_tracking_records = {}
                    self._all_tracking_records[self.application_name] = app_tracking_records
                try:
                    upstream_tracking_records = app_tracking_records[upstream_application_name]
                except KeyError:
                    upstream_tracking_records = set()
                    app_tracking_records[upstream_application_name] = upstream_tracking_records

                if notification_id in upstream_tracking_records:
                    raise RecordConflictError((application_name, upstream_application_name, notification_id))
                upstream_tracking_records.add(notification_id)

    def _insert_records(self, records):
        if isinstance(records, list):
            for record in records:
                self._insert_record(record)
        else:
            self._insert_record(records)

    def _insert_record(self, sequenced_item):
        position = getattr(sequenced_item, self.field_names.position)
        if not isinstance(position, int):
            raise NotImplementedError("Popo record manager only supports sequencing with integers, "
                                      "but position was a {}".format(type(position)))

        sequence_id = getattr(sequenced_item, self.field_names.sequence_id)
        try:
            application_records = self._all_sequence_records[self.application_name]
        except KeyError:
            sequence_records = {}
            application_records = {sequence_id: sequence_records}
            self._all_sequence_records[self.application_name] = application_records
            self._all_sequence_max[self.application_name] = {}
        else:
            try:
                sequence_records = application_records[sequence_id]
            except KeyError:
                sequence_records = {}
                application_records[sequence_id] = sequence_records

        if position in sequence_records:
            raise RecordConflictError(position, len(sequence_records))

        if self.notification_id_name:
            # Just make sure we aren't making a gap in the sequence.
            if sequence_records:
                max_position = self._all_sequence_max[self.application_name][sequence_id]
                next_position = max_position + 1
            else:
                next_position = 0
            if position != next_position:
                raise AssertionError("Next position for sequence {} is {}, not {}".format(
                    sequence_id, next_position, position))

        sequence_records[position] = sequenced_item
        self._all_sequence_max[self.application_name][sequence_id] = position

        # Write a notification record.
        if self.notification_id_name:
            try:
                notification_records = self._all_notification_records[self.application_name]
            except KeyError:
                notification_records = {}
                self._all_notification_records[self.application_name] = notification_records

            next_notification_id = (self._get_max_record_id() or 0) + 1
            notification_records[next_notification_id] = {
                'notification_id': next_notification_id,
                'sequenced_item': sequenced_item,
            }
            self._all_notification_max[self.application_name] = next_notification_id

    def to_records(self, sequenced_items):
        return sequenced_items



from eventsourcing.tests.sequenced_item_tests.test_django_record_manager import DjangoTestCase
from eventsourcing.application.django import DjangoApplication
from eventsourcing.tests.test_actors import TestActors


class TestActorsWithDjango(DjangoTestCase, TestActors):
    infrastructure_class = DjangoApplication



from collections import deque

from eventsourcing.domain.model.entity import TimestampedVersionedEntity, EntityWithHashchain


class BaseAggregateRoot(TimestampedVersionedEntity):
    """
    Root entity for an aggregate in a domain driven design.
    """

    class Event(TimestampedVersionedEntity.Event):
        """Supertype for base aggregate root events."""

    class Created(Event, TimestampedVersionedEntity.Created):
        """Triggered when an aggregate root is created."""

    class AttributeChanged(Event, TimestampedVersionedEntity.AttributeChanged):
        """Triggered when an aggregate root attribute is changed."""

    class Discarded(Event, TimestampedVersionedEntity.Discarded):
        """Triggered when an aggregate root is discarded."""

    def __init__(self, **kwargs):
        super(BaseAggregateRoot, self).__init__(**kwargs)
        self.__pending_events__ = deque()

    def __publish__(self, event):
        """
        Overrides super method by adding event
        to internal collection of pending events,
        rather than actually publishing the event.
        """
        self.__pending_events__.append(event)

    def __save__(self):
        """
        Actually publishes all pending events.
        """
        batch_of_events = self.__batch_pending_events__()
        if batch_of_events:
            self.__publish_to_subscribers__(batch_of_events)
            # Don't catch exception and put the events back on the queue.
            # Losing them here is consistent with the behaviour of DomainEntity
            # when an event cannot be stored: the event is effectively lost, the
            # state of the entity must be reset, and the operation repeated.
            # In case of an aggregate sequence conflict, developers need to
            # know what has happened since the last save, so can retry only the
            # command(s) that have caused conflict. Best to save once per command,
            # so the command can be retried cleanly. The purpose of the save
            # method is to allow many events from one command to be persisted
            # together. The save command can be used to persist all the events
            # from many commands, but in case of a failed save after several
            # commands have been executed, it is important to know which
            # commands to retry.

    def __batch_pending_events__(self):
        batch_of_events = []
        try:
            while True:
                batch_of_events.append(self.__pending_events__.popleft())
        except IndexError:
            pass
        return batch_of_events


class AggregateRootWithHashchainedEvents(EntityWithHashchain, BaseAggregateRoot):
    """Extends aggregate root base class with hash-chained events."""

    class Event(EntityWithHashchain.Event, BaseAggregateRoot.Event):
        """Supertype for aggregate events."""

    class Created(Event, EntityWithHashchain.Created, BaseAggregateRoot.Created):
        """Triggered when an aggregate root is created."""

    class AttributeChanged(Event, BaseAggregateRoot.AttributeChanged):
        """Triggered when an aggregate root attribute is changed."""

    class Discarded(Event, EntityWithHashchain.Discarded, BaseAggregateRoot.Discarded):
        """Triggered when an aggregate root is discarded."""


class AggregateRoot(AggregateRootWithHashchainedEvents):
    """Original name for aggregate root base class with hash-chained events."""

    class Event(AggregateRootWithHashchainedEvents.Event):
        """Supertype for aggregate events."""

    class Created(Event, AggregateRootWithHashchainedEvents.Created):
        """Triggered when an aggregate root is created."""

    class AttributeChanged(Event, AggregateRootWithHashchainedEvents.AttributeChanged):
        """Triggered when an aggregate root attribute is changed."""

    class Discarded(Event, AggregateRootWithHashchainedEvents.Discarded):
        """Triggered when an aggregate root is discarded."""



from abc import ABC, abstractmethod

from eventsourcing.exceptions import OperationalError, RecordConflictError
from eventsourcing.infrastructure.sequenceditem import SequencedItem, SequencedItemFieldNames

DEFAULT_PIPELINE_ID = 0


class AbstractSequencedItemRecordManager(ABC):
    def __init__(self, record_class, sequenced_item_class=SequencedItem,
                 contiguous_record_ids=False, application_name=None,
                 pipeline_id=DEFAULT_PIPELINE_ID):
        self.record_class = record_class
        self.sequenced_item_class = sequenced_item_class
        self.field_names = SequencedItemFieldNames(self.sequenced_item_class)

        if hasattr(self.record_class, 'id'):
            self.notification_id_name = 'id'
        elif hasattr(self.record_class, 'notification_id'):
            self.notification_id_name = 'notification_id'
        else:
            self.notification_id_name = ''

        self.contiguous_record_ids = contiguous_record_ids and self.notification_id_name
        if hasattr(self.record_class, 'application_name'):
            assert application_name, "'application_name' not set when required"
            assert contiguous_record_ids, "'contiguous_record_ids' not set when required"
        self.application_name = application_name
        if hasattr(self.record_class, 'pipeline_id'):
            assert hasattr(self.record_class, 'application_name'), "'application_name' column not defined"
        self.pipeline_id = pipeline_id

    def clone(self, application_name, pipeline_id, **kwargs):
        return type(self)(
            record_class=self.record_class,
            contiguous_record_ids=self.contiguous_record_ids,
            sequenced_item_class=self.sequenced_item_class,
            application_name=application_name,
            pipeline_id=pipeline_id,
            **kwargs
        )

    @abstractmethod
    def record_sequenced_items(self, sequenced_item_or_items):
        """
        Writes sequenced item(s) into the datastore.
        """

    def record_sequenced_item(self, sequenced_item):
        """
        Writes sequenced item into the datastore.
        """
        return self.record_sequenced_items(sequenced_item)

    def get_item(self, sequence_id, position):
        """
        Gets sequenced item from the datastore.
        """
        return self.from_record(self.get_record(sequence_id, position))

    @abstractmethod
    def get_record(self, sequence_id, position):
        """
        Gets record at position in sequence.
        """

    def get_items(self, sequence_id, gt=None, gte=None, lt=None, lte=None, limit=None,
                  query_ascending=True, results_ascending=True):
        """
        Returns sequenced item generator.
        """
        records = self.get_records(
            sequence_id=sequence_id,
            gt=gt,
            gte=gte,
            lt=lt,
            lte=lte,
            limit=limit,
            query_ascending=query_ascending,
            results_ascending=results_ascending,

        )
        for item in map(self.from_record, records):
            yield item

    def list_items(self, *args, **kwargs):
        """
        Returns list of sequenced items.
        """
        return list(self.get_items(*args, **kwargs))

    @abstractmethod
    def get_records(self, sequence_id, gt=None, gte=None, lt=None, lte=None, limit=None,
                    query_ascending=True, results_ascending=True):
        """
        Returns records for a sequence.
        """

    def to_record(self, sequenced_item):
        """
        Constructs a record object from given sequenced item object.
        """
        kwargs = self.get_field_kwargs(sequenced_item)
        # Supply application_name, if needed.
        if hasattr(self.record_class, 'application_name'):
            kwargs['application_name'] = self.application_name
        # Supply pipeline_id, if needed.
        if hasattr(self.record_class, 'pipeline_id'):
            kwargs['pipeline_id'] = self.pipeline_id
        return self.record_class(**kwargs)

    def from_record(self, record):
        """
        Constructs and returns a sequenced item object, from given ORM object.
        """
        kwargs = self.get_field_kwargs(record)
        return self.sequenced_item_class(**kwargs)

    @abstractmethod
    def get_notifications(self, start=None, stop=None, *args, **kwargs):
        """
        Returns records sequenced by notification ID, from application, for pipeline, in given range.

        Args 'start' and 'stop' are positions in a zero-based integer sequence.
        """

    @abstractmethod
    def all_sequence_ids(self):
        """
        Returns all sequence IDs.
        """

    def list_sequence_ids(self):
        return list(self.all_sequence_ids())

    @abstractmethod
    def delete_record(self, record):
        """
        Removes permanently given record from the table.
        """

    def get_field_kwargs(self, item):
        return {name: getattr(item, name) for name in self.field_names}

    def raise_sequenced_item_conflict(self):
        msg = "Position already taken in sequence"
        raise RecordConflictError(msg)

    def raise_index_error(self, position):
        raise IndexError("Sequence index out of range: {}".format(position))

    def raise_record_integrity_error(self, e):
        raise RecordConflictError(e)

    def raise_operational_error(self, e):
        raise OperationalError(e)


class ACIDRecordManager(AbstractSequencedItemRecordManager):
    """
    ACID record managers can write tracking records and event records
    in an atomic transaction, needed for atomic processing in process
    applications.
    """
    tracking_record_field_names = [
        'application_name',
        'upstream_application_name',
        'pipeline_id',
        'notification_id',
    ]

    def __init__(self, tracking_record_class=None, *args, **kwargs):
        super(ACIDRecordManager, self).__init__(*args, **kwargs)
        # assert tracking_record_class is not None
        self.tracking_record_class = tracking_record_class

    def clone(self, **kwargs):
        return super(ACIDRecordManager, self).clone(
            tracking_record_class=self.tracking_record_class, **kwargs
        )

    @abstractmethod
    def write_records(self, records, tracking_kwargs=None, orm_objs=None):
        """
        Writes tracking, event and notification records for a process event.
        :param orm_objs:
        """

    @abstractmethod
    def get_max_record_id(self):
        """Return maximum notification ID in pipeline."""

    @abstractmethod
    def get_max_tracking_record_id(self, upstream_application_name):
        """Return maximum tracking record ID for notification from upstream application in pipeline."""

    @abstractmethod
    def has_tracking_record(self, upstream_application_name, pipeline_id, notification_id):
        """
        True if tracking record exists for notification from upstream in pipeline.
        """

    def get_pipeline_and_notification_id(self, sequence_id, position):
        """
        Returns pipeline ID and notification ID for
        event at given position in given sequence.
        """
        # Todo: Optimise query by selecting only two columns, pipeline_id and id (notification ID)?
        record = self.get_record(sequence_id, position)
        notification_id = getattr(record, self.notification_id_name)
        return record.pipeline_id, notification_id


class SQLRecordManager(ACIDRecordManager):
    """
    This is has code common to (extracted from) the SQLAlchemy and Django record managers.

    This makes the subclasses harder to read and probably more brittle. So it might be better
    to inline this with the subclasses, so that each looks more like normal Django or SQLAlchemy
    code. Also, the record manager test cases don't cover the notification log and tracking record
    functionality needed by ProcessApplication, and should so that other record managers can more
    easily be developed.
    """
    def __init__(self, *args, **kwargs):
        super(SQLRecordManager, self).__init__(*args, **kwargs)
        self._insert_select_max = None
        self._insert_values = None
        self._insert_tracking_record = None

    def record_sequenced_items(self, sequenced_item_or_items):
        # Convert sequenced item(s) to database record(s).
        records = self.to_records(sequenced_item_or_items)

        # Write records.
        self.write_records(records)

    def to_records(self, sequenced_item_or_items):
        if isinstance(sequenced_item_or_items, list):
            records = [self.to_record(i) for i in sequenced_item_or_items]
        else:
            records = [self.to_record(sequenced_item_or_items)]
        return records

    @property
    def insert_select_max(self):
        """
        SQL statement that inserts records with contiguous IDs,
        by selecting max ID from indexed table records.
        """
        if self._insert_select_max is None:
            if hasattr(self.record_class, 'application_name'):
                # Todo: Maybe make it support application_name without pipeline_id?
                assert hasattr(self.record_class, 'pipeline_id'), self.record_class
                tmpl = self._insert_select_max_tmpl + self._where_application_name_tmpl
            else:
                tmpl = self._insert_select_max_tmpl
            self._insert_select_max = self._prepare_insert(
                tmpl=tmpl,
                record_class=self.record_class,
                field_names=list(self.field_names),
            )
        return self._insert_select_max

    @abstractmethod
    def _prepare_insert(self, tmpl, record_class, field_names, placeholder_for_id=False):
        """
        Compile SQL statement with placeholders for bind parameters.
        """

    _insert_select_max_tmpl = (
        "INSERT INTO {tablename} ({notification_id}, {columns}) "
        "SELECT COALESCE(MAX({tablename}.{notification_id}), 0) + 1, {placeholders} "
        "FROM ""{tablename}"
    )

    _where_application_name_tmpl = None

    @property
    def insert_values(self):
        """
        SQL statement that inserts records without ID.
        """
        if self._insert_values is None:
            self._insert_values = self._prepare_insert(
                tmpl=self._insert_values_tmpl,
                placeholder_for_id=True,
                record_class=self.record_class,
                field_names=self.field_names,
            )
        return self._insert_values

    @property
    def insert_tracking_record(self):
        """
        SQL statement that inserts tracking records.
        """
        if self._insert_tracking_record is None:
            self._insert_tracking_record = self._prepare_insert(
                tmpl=self._insert_values_tmpl,
                placeholder_for_id=True,
                record_class=self.tracking_record_class,
                field_names=self.tracking_record_field_names,
            )
        return self._insert_tracking_record

    _insert_values_tmpl = (
        "INSERT INTO {tablename} ({columns}) "
        "VALUES ({placeholders});"
    )

    @abstractmethod
    def get_record_table_name(self, record_class):
        """
        Returns table name - used in raw queries.

        :rtype: str
        """



from eventsourcing.tests.sequenced_item_tests.test_django_record_manager import DjangoTestCase
from eventsourcing.tests.contrib_tests.paxos_tests import test_paxos_system
from eventsourcing.application.django import DjangoApplication


class TestPaxosSystemWithDjango(DjangoTestCase, test_paxos_system.TestPaxosSystem):
    infrastructure_class = DjangoApplication



# from eventsourcing.application.sqlalchemy import SQLAlchemyApplication
# from eventsourcing.domain.model.aggregate import AggregateRoot
# # from player import Player
#
#
# db_path = 'sqlite:////tmp/local_db'
#
# application = SQLAlchemyApplication(
#     uri=db_path,
#     persist_event_type=AggregateRoot.Event
# )
#
# player_1 = AggregateRoot.__create__()
#
# print(f"Player ID: {player_1.id}" )
#
# player_1.__save__()
#
# same_p = application.repository[player_1.id]
# print(f"Retrieved ID: {same_p.id}")
# print(f"Retrieved ID: {type(same_p.id)}")



from tempfile import NamedTemporaryFile
from uuid import uuid4

from sqlalchemy.exc import OperationalError, ProgrammingError

from eventsourcing.infrastructure.datastore import DatastoreTableError
from eventsourcing.infrastructure.sqlalchemy.datastore import DEFAULT_SQLALCHEMY_DB_URI, SQLAlchemyDatastore, \
    SQLAlchemySettings
from eventsourcing.infrastructure.sqlalchemy.factory import SQLAlchemyInfrastructureFactory
from eventsourcing.infrastructure.sqlalchemy.records import Base, IntegerSequencedNoIDRecord, \
    IntegerSequencedWithIDRecord, SnapshotRecord, TimestampSequencedNoIDRecord, TimestampSequencedWithIDRecord
from eventsourcing.tests.datastore_tests import base


class SQLAlchemyDatastoreTestCase(base.AbstractDatastoreTestCase):
    """
    Base class for test cases that use an SQLAlchemy datastore.
    """

    use_named_temporary_file = False
    connection_strategy = 'plain'
    infrastructure_factory_class = SQLAlchemyInfrastructureFactory
    contiguous_record_ids = True

    def construct_datastore(self):
        if self.use_named_temporary_file:
            self.temp_file = NamedTemporaryFile('a', delete=True)
            uri = 'sqlite:///' + self.temp_file.name
        else:
            uri = DEFAULT_SQLALCHEMY_DB_URI

        # kwargs = {}
        # if not self.use_named_temporary_file:
        # kwargs['connect_args'] = {'check_same_thread':False}
        # kwargs['poolclass'] = StaticPool

        return SQLAlchemyDatastore(
            base=Base,
            settings=SQLAlchemySettings(uri=uri),
            tables=(
                IntegerSequencedWithIDRecord,
                IntegerSequencedNoIDRecord,
                TimestampSequencedWithIDRecord,
                TimestampSequencedNoIDRecord,
                SnapshotRecord
            ),
            connection_strategy=self.connection_strategy,
            # **kwargs
        )


class TestSQLAlchemyDatastore(SQLAlchemyDatastoreTestCase, base.DatastoreTestCase):
    """
    Test case for SQLAlchemy datastore.
    """
    def list_records(self):
        try:
            query = self.datastore.session.query(IntegerSequencedNoIDRecord)
            return list(query)
        except (OperationalError, ProgrammingError) as e:
            # OperationalError from sqlite, ProgrammingError from psycopg2.
            self.datastore.session.rollback()
            raise DatastoreTableError(e)
        finally:
            self.datastore.session.close()

    def create_record(self):
        try:
            record = IntegerSequencedNoIDRecord(
                sequence_id=uuid4(),
                position=0,
                topic='topic',
                state='{}'
            )
            self.datastore.session.add(record)
            self.datastore.session.commit()
        except (OperationalError, ProgrammingError) as e:
            self.datastore.session.rollback()
            raise DatastoreTableError(e)
        return record



import random
from functools import singledispatch, wraps
from inspect import isfunction
from time import sleep

from eventsourcing.domain.model.events import subscribe
from eventsourcing.exceptions import ProgrammingError


def subscribe_to(*event_classes):
    """
    Decorator for making a custom event handler function subscribe to a certain class of event.

    The decorated function will be called once for each matching event that is published, and will
    be given one argument, the event, when it is called. If events are published in lists, for
    example the AggregateRoot publishes a list of pending events when its __save__() method is called,
    then the decorated function will be called once for each event that is an instance of the given event_class.

    Please note, this decorator isn't suitable for use with object class methods. The decorator receives
    in Python 3 an unbound function, and defines a handler which it subscribes that calls the decorated
    function for each matching event. However the method isn't called on the object, so the object instance
    is never available in the decorator, so the decorator can't call a normal object method because it
    doesn't have a value for 'self'.

    event_class: type used to match published events, an event matches if it is an instance of this type

    The following example shows a custom handler that reacts to Todo.Created
    event and saves a projection of a Todo model object.

    .. code::

        @subscribe_to(Todo.Created)
        def new_todo_projection(event):
            todo = TodoProjection(id=event.originator_id, title=event.title)
            todo.save()
    """
    event_classes = list(event_classes)

    def wrap(func):
        def handler(event):
            if isinstance(event, (list, tuple)):
                for e in event:
                    handler(e)
            elif not event_classes or isinstance(event, tuple(event_classes)):
                func(event)

        subscribe(handler=handler, predicate=lambda _: True)
        return func

    if len(event_classes) == 1 and isfunction(event_classes[0]):
        func = event_classes.pop()
        return wrap(func)
    else:
        return wrap


def mutator(arg=None):
    """Structures mutator functions by allowing handlers
    to be registered for different types of event. When
    the decorated function is called with an initial
    value and an event, it will call the handler that
    has been registered for that type of event.

    It works like singledispatch, which it uses. The
    difference is that when the decorated function is
    called, this decorator dispatches according to the
    type of last call arg, which fits better with reduce().
    The builtin Python function reduce() is used by the
    library to replay a sequence of events against an
    initial state. If a mutator function is given to reduce(),
    along with a list of events and an initializer, reduce()
    will call the mutator function once for each event in the
    list, but the initializer will be the first value, and the
    event will be the last argument, and we want to dispatch
    according to the type of the event. It happens that
    singledispatch is coded to switch on the type of the first
    argument, which makes it unsuitable for structuring a mutator
    function without the modifications introduced here.

    The other aspect introduced by this decorator function is the
    option to set the type of the handled entity in the decorator.
    When an entity is replayed from scratch, in other words when
    all its events are replayed, the initial state is None. The
    handler which handles the first event in the sequence will
    probably construct an object instance. It is possible to write
    the type into the handler, but that makes the entity more difficult
    to subclass because you will also need to write a handler for it.
    If the decorator is invoked with the type, when the initial
    value passed as a call arg to the mutator function is None,
    the handler will instead receive the type of the entity, which
    it can use to construct the entity object.

    .. code::

        class Entity(object):
            class Created(object):
                pass

        @mutator(Entity)
        def mutate(initial, event):
            raise NotImplementedError(type(event))

        @mutate.register(Entity.Created)
        def _(initial, event):
            return initial(**event.__dict__)

        entity = mutate(None, Entity.Created())
    """

    domain_class = None

    def _mutator(func):
        wrapped = singledispatch(func)

        @wraps(wrapped)
        def wrapper(initial, event):
            initial = initial or domain_class
            return wrapped.dispatch(type(event))(initial, event)

        wrapper.register = wrapped.register

        return wrapper

    if isfunction(arg):
        return _mutator(arg)
    else:
        domain_class = arg
        return _mutator


def attribute(getter):
    """
    When used as a method decorator, returns a property object
    with the method as the getter and a setter defined to call
    instance method __change_attribute__(), which publishes an
    AttributeChanged event.
    """
    if isfunction(getter):
        def setter(self, value):
            name = '_' + getter.__name__
            self.__change_attribute__(name=name, value=value)

        def new_getter(self):
            name = '_' + getter.__name__
            return getattr(self, name, None)

        return property(fget=new_getter, fset=setter, doc=getter.__doc__)
    else:
        raise ProgrammingError("Expected a function, got: {}".format(repr(getter)))


def retry(exc=Exception, max_attempts=1, wait=0, stall=0, verbose=False):
    """
    Retry decorator.

    :param exc: List of exceptions that will cause the call to be retried if raised.
    :param max_attempts: Maximum number of attempts to try.
    :param wait: Amount of time to wait before retrying after an exception.
    :param stall: Amount of time to wait before the first attempt.
    :param verbose: If True, prints a message to STDOUT when retries occur.
    :return: Returns the value returned by decorated function.
    """
    def _retry(func):

        @wraps(func)
        def wrapper(*args, **kwargs):
            if stall:
                sleep(stall)
            attempts = 0
            while True:
                try:
                    return func(*args, **kwargs)
                except exc as e:
                    attempts += 1
                    if max_attempts is None or attempts < max_attempts:
                        sleep(wait * (1 + 0.1 * (random.random() - 0.5)))
                        if verbose:
                            print("Retrying {}".format(func))
                    else:
                        # Max retries exceeded.
                        raise e

        return wrapper

    # If using decorator in bare form, the decorated
    # function is the first arg, so check 'exc'.
    if isfunction(exc):
        # Remember the given function.
        _func = exc
        # Set 'exc' to a sensible exception class for _retry().
        exc = Exception
        # Wrap and return.
        return _retry(func=_func)
    else:
        # Check decorator args, and return _retry,
        # to be called with the decorated function.
        if isinstance(exc, (list, tuple)):
            for _exc in exc:
                if not (isinstance(_exc, type) and issubclass(_exc, Exception)):
                    raise TypeError("not an exception class: {}".format(_exc))
        else:
            if not (isinstance(exc, type) and issubclass(exc, Exception)):
                raise TypeError("not an exception class: {}".format(exc))
        if not isinstance(max_attempts, int):
            raise TypeError("'max_attempts' must be an int: {}".format(max_attempts))
        if not isinstance(wait, (float, int)):
            raise TypeError("'wait' must be a float: {}".format(max_attempts))
        return _retry



# coding=utf-8
from __future__ import unicode_literals

import datetime
from uuid import uuid4

from eventsourcing.domain.model.decorators import attribute
from eventsourcing.domain.model.entity import AttributeChanged, Created, Discarded, TimestampedVersionedEntity
from eventsourcing.domain.model.events import publish
from eventsourcing.example.application import ExampleApplication
from eventsourcing.exceptions import RepositoryKeyError
from eventsourcing.infrastructure.eventsourcedrepository import EventSourcedRepository


class SuffixTree(TimestampedVersionedEntity):
    """A suffix tree for string matching. Uses Ukkonen's algorithm
    for construction.
    """

    class Created(Created):
        pass

    class AttributeChanged(AttributeChanged):
        pass

    class Discarded(Discarded):
        pass

    def __init__(self, root_node_id, case_insensitive=False, **kwargs):
        """
        string
            the string for which to construct a suffix tree
        """
        super(SuffixTree, self).__init__(**kwargs)
        self._root_node_id = root_node_id
        self._case_insensitive = case_insensitive
        self._nodes = {}
        self._edges = {}
        self._N = None
        self._string = None

    @attribute
    def string(self):
        return self._string

    @property
    def N(self):
        return self._N

    @property
    def nodes(self):
        return self._nodes

    @property
    def root_node_id(self):
        return self._root_node_id

    @property
    def edges(self):
        return self._edges

    @property
    def active(self):
        return self._active

    @property
    def case_insensitive(self):
        return self._case_insensitive

    def __repr__(self):
        """
        Lists edges in the suffix tree
        """
        return "SuffixTree({})".format(self.id)

    def add_string(self, string):
        assert self.string is None
        self._active = Suffix(self._root_node_id, 0, -1)
        self._N = len(string) - 1
        if self._case_insensitive:
            string = string.lower()
        self.string = string
        for i in range(len(string)):
            self._add_prefix(i)

    def _add_prefix(self, last_char_index):
        """The core construction method.
        """
        last_parent_node_id = None
        while True:
            parent_node_id = self.active.source_node_id
            if self.active.explicit():
                edge_id = make_edge_id(self.active.source_node_id, self.string[last_char_index])
                if edge_id in self.edges:
                    # prefix is already in tree
                    break
            else:
                edge_id = make_edge_id(self.active.source_node_id, self.string[self.active.first_char_index])
                e = self.edges[edge_id]
                if self.string[e.first_char_index + self.active.length + 1] == self.string[last_char_index]:
                    # prefix is already in tree
                    break
                parent_node_id = self._split_edge(e, self.active)

            node = register_new_node()
            self.nodes[node.id] = node
            edge_id = make_edge_id(parent_node_id, self.string[last_char_index])
            e = register_new_edge(
                edge_id=edge_id,
                first_char_index=last_char_index,
                last_char_index=self.N,
                source_node_id=parent_node_id,
                dest_node_id=node.id,
            )
            self._insert_edge(e)

            if last_parent_node_id is not None:
                self.nodes[last_parent_node_id].suffix_node_id = parent_node_id
            last_parent_node_id = parent_node_id

            if self.active.source_node_id == self.root_node_id:
                self.active.first_char_index += 1
            else:
                self.active.source_node_id = self.nodes[self.active.source_node_id].suffix_node_id
            self._canonize_suffix(self.active)
        if last_parent_node_id is not None:
            self.nodes[last_parent_node_id].suffix_node_id = parent_node_id
        self.active.last_char_index += 1
        self._canonize_suffix(self.active)

    def _insert_edge(self, edge):
        edge_id = make_edge_id(edge.source_node_id, self.string[edge.first_char_index])
        self.edges[edge_id] = edge

    def _remove_edge(self, edge):
        edge_id = make_edge_id(edge.source_node_id, self.string[edge.first_char_index])
        self.edges.pop(edge_id)

    def _split_edge(self, first_edge, suffix):
        assert isinstance(first_edge, Edge)
        assert isinstance(suffix, Suffix)

        # Create a middle node that will split the edge.
        new_node = register_new_node(suffix.source_node_id)
        self.nodes[new_node.id] = new_node

        # Split the char indexes.
        first_edge_last_char_index = first_edge.first_char_index + suffix.length
        second_edge_first_char_index = first_edge.first_char_index + suffix.length + 1
        second_edge_last_char_index = first_edge.last_char_index

        # Create a new edge, from the middle node to the original destination.
        second_edge_id = make_edge_id(new_node.id, self.string[first_edge.first_char_index + suffix.length + 1])
        second_edge = register_new_edge(
            edge_id=second_edge_id,
            first_char_index=second_edge_first_char_index,
            last_char_index=second_edge_last_char_index,
            source_node_id=new_node.id,
            dest_node_id=first_edge.dest_node_id,
        )
        self._insert_edge(second_edge)

        # Shorten the first edge.
        first_edge.last_char_index = first_edge_last_char_index
        first_edge.dest_node_id = new_node.id

        # Return middle node.
        return new_node.id

    def _canonize_suffix(self, suffix):
        """This canonizes the suffix, walking along its suffix string until it
        is explicit or there are no more matched nodes.
        """
        if not suffix.explicit():
            edge_id = make_edge_id(suffix.source_node_id, self.string[suffix.first_char_index])
            e = self.edges[edge_id]
            if e.length <= suffix.length:
                suffix.first_char_index += e.length + 1
                suffix.source_node_id = e.dest_node_id
                self._canonize_suffix(suffix)


class Node(TimestampedVersionedEntity):
    """A node in the suffix tree.
    """

    class Created(Created): pass

    class AttributeChanged(AttributeChanged): pass

    class Discarded(Discarded): pass

    def __init__(self, suffix_node_id=None, *args, **kwargs):
        super(Node, self).__init__(*args, **kwargs)
        self._suffix_node_id = suffix_node_id

    @attribute
    def suffix_node_id(self):
        """The id of a node with a matching suffix, representing a suffix link.

        None indicates this node has no suffix link.
        """
        return self._suffix_node_id

    def __repr__(self):
        return "Node(suffix link: %d)" % self.suffix_node_id


class Edge(TimestampedVersionedEntity):
    """An edge in the suffix tree.
    """

    class Created(Created): pass

    class AttributeChanged(AttributeChanged): pass

    class Discarded(Discarded): pass

    def __init__(self, first_char_index, last_char_index, source_node_id, dest_node_id, **kwargs):
        super(Edge, self).__init__(**kwargs)
        self._first_char_index = first_char_index
        self._last_char_index = last_char_index
        self._source_node_id = source_node_id
        self._dest_node_id = dest_node_id

    @attribute
    def first_char_index(self):
        """Index of start of string part represented by this edge.
        """
        return self._first_char_index

    @attribute
    def last_char_index(self):
        """Index of end of string part represented by this edge.
        """
        return self._last_char_index

    @attribute
    def source_node_id(self):
        """Id of source node of edge.
        """
        return self._source_node_id

    @attribute
    def dest_node_id(self):
        """Id of destination node of edge.
        """
        return self._dest_node_id

    @property
    def length(self):
        """Number of chars in the string part represented by this edge.
        """
        return self.last_char_index - self.first_char_index

    def __repr__(self):
        return 'Edge(%d, %d, %d, %d)' % (self.source_node_id, self.dest_node_id
                                         , self.first_char_index, self.last_char_index)


class Suffix(object):
    """Represents a suffix from first_char_index to last_char_index.
    """

    def __init__(self, source_node_id, first_char_index, last_char_index):
        self._source_node_id = source_node_id
        self._first_char_index = first_char_index
        self._last_char_index = last_char_index

    @property
    def source_node_id(self):
        """Index of node where this suffix starts.
        """
        return self._source_node_id

    @source_node_id.setter
    def source_node_id(self, value):
        self._source_node_id = value

    @property
    def first_char_index(self):
        """Index of start of suffix in string.
        """
        return self._first_char_index

    @first_char_index.setter
    def first_char_index(self, value):
        self._first_char_index = value

    @property
    def last_char_index(self):
        """Index of end of suffix in string.
        """
        return self._last_char_index

    @last_char_index.setter
    def last_char_index(self, value):
        self._last_char_index = value

    @property
    def length(self):
        """Number of chars in string.
        """
        return self.last_char_index - self.first_char_index

    def explicit(self):
        """A suffix is explicit if it ends on a node. first_char_index
        is set greater than last_char_index to indicate this.
        """
        return self.first_char_index > self.last_char_index

    def implicit(self):
        return self.last_char_index >= self.first_char_index


# Factory methods.

def register_new_node(suffix_node_id=None):
    """Factory method, registers new node.
    """
    node_id = uuid4()
    event = Node.Created(originator_id=node_id, suffix_node_id=suffix_node_id)
    entity = Node.mutate(event=event)
    publish(event)
    return entity


def make_edge_id(source_node_index, first_char):
    """Returns a string made from given params.
    """
    return "{}::{}".format(source_node_index, first_char)


def register_new_edge(edge_id, first_char_index, last_char_index, source_node_id, dest_node_id):
    """Factory method, registers new edge.
    """
    event = Edge.Created(
        originator_id=edge_id,
        first_char_index=first_char_index,
        last_char_index=last_char_index,
        source_node_id=source_node_id,
        dest_node_id=dest_node_id,
    )
    entity = Edge.mutate(event=event)
    publish(event)
    return entity


def register_new_suffix_tree(case_insensitive=False):
    """Factory method, returns new suffix tree object.
    """
    assert isinstance(case_insensitive, bool)
    root_node = register_new_node()

    suffix_tree_id = uuid4()
    event = SuffixTree.Created(
        originator_id=suffix_tree_id,
        root_node_id=root_node.id,
        case_insensitive=case_insensitive,
    )
    entity = SuffixTree.mutate(event=event)

    assert isinstance(entity, SuffixTree)

    entity.nodes[root_node.id] = root_node

    publish(event)

    return entity


# Domain services

def find_substring(substring, suffix_tree, edge_repo):
    """Returns the index if substring in tree, otherwise -1.
    """
    assert isinstance(substring, str)
    assert isinstance(suffix_tree, SuffixTree)
    assert isinstance(edge_repo, EventSourcedRepository)
    if not substring:
        return -1
    if suffix_tree.case_insensitive:
        substring = substring.lower()
    curr_node_id = suffix_tree.root_node_id
    i = 0
    while i < len(substring):
        edge_id = make_edge_id(curr_node_id, substring[i])
        try:
            edge = edge_repo[edge_id]
        except RepositoryKeyError:
            return -1
        ln = min(edge.length + 1, len(substring) - i)
        if substring[i:i + ln] != suffix_tree.string[edge.first_char_index:edge.first_char_index + ln]:
            return -1
        i += edge.length + 1
        curr_node_id = edge.dest_node_id
    return edge.first_char_index - len(substring) + ln


def has_substring(substring, suffix_tree, edge_repo):
    return find_substring(substring, suffix_tree, edge_repo) != -1


class SuffixTreeApplication(ExampleApplication):

    def register_new_suffixtree(self, case_insensitive=False):
        return register_new_suffix_tree(case_insensitive)

    def find_substring(self, substring, suffix_tree_id):
        suffix_tree = self.example_repository[suffix_tree_id]
        started = datetime.datetime.now()
        result = find_substring(substring=substring, suffix_tree=suffix_tree, edge_repo=self.example_repository)
        print("- found substring '{}' in: {}".format(substring, datetime.datetime.now() - started))
        return result

    def has_substring(self, substring, suffix_tree_id):
        suffix_tree = self.example_repository[suffix_tree_id]
        return has_substring(
            substring=substring,
            suffix_tree=suffix_tree,
            edge_repo=self.example_repository,
        )



from eventsourcing.infrastructure.base import DEFAULT_PIPELINE_ID
from eventsourcing.infrastructure.eventstore import EventStore
from eventsourcing.infrastructure.factory import InfrastructureFactory
from eventsourcing.infrastructure.sequenceditem import StoredEvent
from eventsourcing.infrastructure.sequenceditemmapper import SequencedItemMapper
from eventsourcing.infrastructure.sqlalchemy.datastore import SQLAlchemyDatastore, SQLAlchemySettings
from eventsourcing.infrastructure.sqlalchemy.manager import SQLAlchemyRecordManager
from eventsourcing.infrastructure.sqlalchemy.records import IntegerSequencedWithIDRecord, SnapshotRecord, \
    StoredEventRecord, TimestampSequencedNoIDRecord, NotificationTrackingRecord


class SQLAlchemyInfrastructureFactory(InfrastructureFactory):
    """
    Infrastructure factory for SQLAlchemy infrastructure.
    """
    record_manager_class = SQLAlchemyRecordManager
    integer_sequenced_record_class = IntegerSequencedWithIDRecord
    timestamp_sequenced_record_class = TimestampSequencedNoIDRecord
    snapshot_record_class = SnapshotRecord
    tracking_record_class = NotificationTrackingRecord

    def __init__(self, session, uri=None, pool_size=None, tracking_record_class=None, *args, **kwargs):
        super(SQLAlchemyInfrastructureFactory, self).__init__(*args, **kwargs)
        self.session = session
        self.uri = uri
        self.pool_size = pool_size
        self._tracking_record_class = tracking_record_class

    def construct_integer_sequenced_record_manager(self, **kwargs):
        """
        Constructs SQLAlchemy record manager.

        :return: An SQLAlchemy record manager.
        :rtype: SQLAlchemyRecordManager
        """
        tracking_record_class = self._tracking_record_class or self.tracking_record_class
        return super(SQLAlchemyInfrastructureFactory, self).construct_integer_sequenced_record_manager(
            tracking_record_class=tracking_record_class,
            **kwargs
        )

    def construct_record_manager(self, record_class, **kwargs):
        """
        Constructs SQLAlchemy record manager.

        :return: An SQLAlchemy record manager.
        :rtype: SQLAlchemyRecordManager
        """
        return super(SQLAlchemyInfrastructureFactory, self).construct_record_manager(
            record_class,
            session=self.session,
            **kwargs
        )

    def construct_datastore(self):
        """
        Constructs SQLAlchemy datastore.

        :rtype: SQLAlchemyDatastore
        """
        datastore = SQLAlchemyDatastore(
            settings=SQLAlchemySettings(
                uri=self.uri,
                pool_size=self.pool_size
            ),
            session=self.session,
        )
        self.session = datastore.session
        return datastore


def construct_sqlalchemy_eventstore(session,
                                    sequenced_item_class=None,
                                    sequence_id_attr_name=None,
                                    position_attr_name=None,
                                    json_encoder_class=None,
                                    json_decoder_class=None,
                                    cipher=None,
                                    record_class=None,
                                    contiguous_record_ids=False,
                                    application_name=None,
                                    pipeline_id=DEFAULT_PIPELINE_ID,
                                    ):
    sequenced_item_class = sequenced_item_class or StoredEvent
    sequenced_item_mapper = SequencedItemMapper(
        sequenced_item_class=sequenced_item_class,
        sequence_id_attr_name=sequence_id_attr_name,
        position_attr_name=position_attr_name,
        json_encoder_class=json_encoder_class,
        json_decoder_class=json_decoder_class,
        cipher=cipher,
    )
    factory = SQLAlchemyInfrastructureFactory(
        session=session,
        integer_sequenced_record_class=record_class or StoredEventRecord,
        sequenced_item_class=sequenced_item_class,
        contiguous_record_ids=contiguous_record_ids,
        application_name=application_name,
        pipeline_id=pipeline_id,
    )
    record_manager = factory.construct_integer_sequenced_record_manager()
    event_store = EventStore(
        record_manager=record_manager,
        sequenced_item_mapper=sequenced_item_mapper,
    )
    return event_store



from unittest import TestCase

from eventsourcing.infrastructure.sequenceditem import SequencedItem


class TestIntegerSequencedItem(TestCase):
    def test(self):
        sequence_id = 'sequence1'
        position = 0
        topic = 'topic1'
        state = '{}'
        item = SequencedItem(
            sequence_id=sequence_id,
            position=position,
            topic=topic,
            state=state,
        )
        self.assertEqual(item.sequence_id, sequence_id)
        self.assertEqual(item.position, position)
        self.assertEqual(item.topic, topic)
        self.assertEqual(item.state, state)

        with self.assertRaises(AttributeError):
            item.sequence_id = 'sequence2'


class TestTimeSequencedItem(TestCase):
    def test(self):
        sequence_id = 'sequence1'
        position = 0
        topic = 'topic1'
        state = '{}'
        item = SequencedItem(
            sequence_id=sequence_id,
            position=position,
            topic=topic,
            state=state,
        )
        self.assertEqual(item.sequence_id, sequence_id)
        self.assertEqual(item.position, position)
        self.assertEqual(item.topic, topic)
        self.assertEqual(item.state, state)

        with self.assertRaises(AttributeError):
            item.sequence_id = 'sequence2'



from eventsourcing.application.process import ProcessApplication
from eventsourcing.domain.model.command import Command


class CommandProcess(ProcessApplication):
    persist_event_type = Command.Event



"""
This module provides an implementation of the Paxos algorithm as
a set of composable classes.

Copied, with minor alterations for use with Python 3 and JSON from
https://github.com/cocagne/python-composable-paxos

The MIT License (MIT)

Copyright (c) 2015 Tom Cocagne

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

# ProposalID
#
# In order for the Paxos algorithm to function, all proposal ids must be
# unique. A simple way to ensure this is to include the proposer's unique
# id in the proposal id.
#
# Python tuples allow the proposal number and the UID to be combined in a
# manner that supports comparison in the expected manner:
#
#   (4,'C') > (4,'B') > (3,'Z')
#
# Named tuples from the collections module support all of the regular
# tuple operations but additionally allow access to the contents by
# name so the numeric component of the proposal ID may be referred to
# via 'proposal_id.number' instead of 'proposal_id[0]'.
#
# ProposalID = collections.namedtuple('ProposalID', ['number', 'uid'])

# Replaced ProposalID class, because json package turns namedtuples into a list.


class ProposalID(object):
    def __init__(self, number, uid):
        self.number = number
        self.uid = uid

    def __gt__(self, other):
        if other is None:
            return True
        elif isinstance(other, ProposalID):
            return (self.number, self.uid) > (other.number, other.uid)
        elif isinstance(other, list):
            return [self.number, self.uid] > other

    def __ge__(self, other):
        if other is None:
            return True
        elif isinstance(other, ProposalID):
            return (self.number, self.uid) >= (other.number, other.uid)
        elif isinstance(other, list):
            return [self.number, self.uid] >= other

    def __eq__(self, other):
        if other is None:
            return False
        elif isinstance(other, ProposalID):
            return (self.number, self.uid) == (other.number, other.uid)
        elif isinstance(other, list):
            return [self.number, self.uid] == other

    def __repr__(self):
        return 'ProposalID(number={}, uid="{}")'.format(self.number, self.uid)

    def __hash__(self):
        return hash((self.number, self.uid))


class PaxosMessage(object):
    '''
    Base class for all messages defined in this module
    '''
    from_uid = None  # Set by subclass constructor


class Prepare(PaxosMessage):
    '''
    Prepare messages should be broadcast to all Acceptors.
    '''

    def __init__(self, from_uid, proposal_id):
        self.from_uid = from_uid
        self.proposal_id = proposal_id


class Nack(PaxosMessage):
    '''
    NACKs are technically optional though few practical applications will
    want to omit their use. They are used to signal a proposer that their
    current proposal number is out of date and that a new one should be
    chosen. NACKs may be sent in response to both Prepare and Accept
    messages
    '''

    def __init__(self, from_uid, proposer_uid, proposal_id, promised_proposal_id):
        self.from_uid = from_uid
        self.proposal_id = proposal_id
        self.proposer_uid = proposer_uid
        self.promised_proposal_id = promised_proposal_id


class Promise(PaxosMessage):
    '''
    Promise messages should be sent to at least the Proposer specified in
    the proposer_uid field
    '''

    def __init__(self, from_uid, proposer_uid, proposal_id, last_accepted_id, last_accepted_value):
        self.from_uid = from_uid
        self.proposer_uid = proposer_uid
        self.proposal_id = proposal_id
        self.last_accepted_id = last_accepted_id
        self.last_accepted_value = last_accepted_value


class Accept(PaxosMessage):
    '''
    Accept messages should be broadcast to all Acceptors
    '''

    def __init__(self, from_uid, proposal_id, proposal_value):
        self.from_uid = from_uid
        self.proposal_id = proposal_id
        self.proposal_value = proposal_value


class Accepted(PaxosMessage):
    '''
    Accepted messages should be sent to all Learners
    '''

    def __init__(self, from_uid, proposal_id, proposal_value):
        self.from_uid = from_uid
        self.proposal_id = proposal_id
        self.proposal_value = proposal_value


class Resolution(PaxosMessage):
    '''
    Optional message used to indicate that the final value has been selected
    '''

    def __init__(self, from_uid, value):
        self.from_uid = from_uid
        self.value = value


class InvalidMessageError(Exception):
    '''
    Thrown if a PaxosMessage subclass is passed to a class that does not
    support it
    '''


class MessageHandler(object):

    def receive(self, msg):
        '''
        Message dispatching function. This function accepts any PaxosMessage subclass and calls
        the appropriate handler function
        '''
        handler = getattr(self, 'receive_' + msg.__class__.__name__.lower(), None)
        if handler is None:
            raise InvalidMessageError('Receiving class does not support messages of type: ' + msg.__class__.__name__)
        return handler(msg)


class Proposer(MessageHandler):
    '''
    The 'leader' attribute is a boolean value indicating the Proposer's
    belief in whether or not it is the current leader. This is not a reliable
    value as multiple nodes may simultaneously believe themselves to be the
    leader.
    '''

    leader = False
    proposed_value = None
    proposal_id = None
    highest_accepted_id = None
    promises_received = None
    nacks_received = None
    current_prepare_msg = None
    current_accept_msg = None

    def __init__(self, network_uid, quorum_size):
        self.network_uid = network_uid
        self.quorum_size = quorum_size
        self.proposal_id = ProposalID(0, network_uid)
        self.highest_proposal_id = ProposalID(0, network_uid)

    def propose_value(self, value):
        '''
        Sets the proposal value for this node iff this node is not already aware of
        a previous proposal value. If the node additionally believes itself to be
        the current leader, an Accept message will be returned
        '''
        if self.proposed_value is None:
            self.proposed_value = value

            if self.leader:
                self.current_accept_msg = Accept(self.network_uid, self.proposal_id, value)
                return self.current_accept_msg

    def prepare(self):
        '''
        Returns a new Prepare message with a proposal id higher than
        that of any observed proposals. A side effect of this method is
        to clear the leader flag if it is currently set.
        '''

        self.leader = False
        self.promises_received = set()
        self.nacks_received = set()
        self.proposal_id = ProposalID(self.highest_proposal_id.number + 1, self.network_uid)
        self.highest_proposal_id = self.proposal_id
        self.current_prepare_msg = Prepare(self.network_uid, self.proposal_id)

        return self.current_prepare_msg

    def observe_proposal(self, proposal_id):
        '''
        Optional method used to update the proposal counter as proposals are
        seen on the network.  When co-located with Acceptors and/or Learners,
        this method may be used to avoid a message delay when attempting to
        assume leadership (guaranteed NACK if the proposal number is too low).
        This method is automatically called for all received Promise and Nack
        messages.
        '''
        if proposal_id > self.highest_proposal_id:
            self.highest_proposal_id = proposal_id

    def receive_nack(self, msg):
        '''
        Returns a new Prepare message if the number of Nacks received reaches
        a quorum.
        '''
        self.observe_proposal(msg.promised_proposal_id)

        if msg.proposal_id == self.proposal_id and self.nacks_received is not None:
            self.nacks_received.add(msg.from_uid)

            if len(self.nacks_received) == self.quorum_size:
                return self.prepare()  # Lost leadership or failed to acquire it

    def receive_promise(self, msg):
        '''
        Returns an Accept messages if a quorum of Promise messages is achieved
        '''
        self.observe_proposal(msg.proposal_id)

        if not self.leader and msg.proposal_id == self.proposal_id and msg.from_uid not in self.promises_received:

            self.promises_received.add(msg.from_uid)

            if self.highest_accepted_id is None or msg.last_accepted_id > self.highest_accepted_id:
                self.highest_accepted_id = msg.last_accepted_id
                if msg.last_accepted_value is not None:
                    self.proposed_value = msg.last_accepted_value

            if len(self.promises_received) == self.quorum_size:
                self.leader = True

                if self.proposed_value is not None:
                    self.current_accept_msg = Accept(self.network_uid, self.proposal_id, self.proposed_value)
                    return self.current_accept_msg


class Acceptor(MessageHandler):
    '''
    Acceptors act as the fault-tolerant memory for Paxos. To ensure correctness
    in the presense of failure, Acceptors must be able to remember the promises
    they've made even in the event of power outages. Consequently, any changes
    to the promised_id, accepted_id, and/or accepted_value must be persisted to
    stable media prior to sending promise and accepted messages.

    When an Acceptor instance is composed alongside a Proposer instance, it
    is generally advantageous to call the proposer's observe_proposal()
    method when methods of this class are called.
    '''

    def __init__(self, network_uid, promised_id=None, accepted_id=None, accepted_value=None):
        '''
        promised_id, accepted_id, and accepted_value should be provided if and only if this
        instance is recovering from persistent state.
        '''
        self.network_uid = network_uid
        self.promised_id = promised_id
        self.accepted_id = accepted_id
        self.accepted_value = accepted_value

    def receive_prepare(self, msg):
        '''
        Returns either a Promise or a Nack in response. The Acceptor's state must be persisted to disk
        prior to transmitting the Promise message.
        '''
        if self.promised_id is None or msg.proposal_id >= self.promised_id:
            self.promised_id = msg.proposal_id
            return Promise(self.network_uid, msg.from_uid, self.promised_id, self.accepted_id, self.accepted_value)
        else:
            return Nack(self.network_uid, msg.from_uid, msg.proposal_id, self.promised_id)

    def receive_accept(self, msg):
        '''
        Returns either an Accepted or Nack message in response. The Acceptor's state must be persisted
        to disk prior to transmitting the Accepted message.
        '''
        if self.promised_id is None or msg.proposal_id >= self.promised_id:
            self.promised_id = msg.proposal_id
            self.accepted_id = msg.proposal_id
            self.accepted_value = msg.proposal_value
            return Accepted(self.network_uid, msg.proposal_id, msg.proposal_value)
        else:
            return Nack(self.network_uid, msg.from_uid, msg.proposal_id, self.promised_id)


class ProposalStatus(object):
    __slots__ = ['accept_count', 'retain_count', 'acceptors', 'value']

    def __init__(self, value):
        self.accept_count = 0
        self.retain_count = 0
        self.acceptors = set()
        self.value = value


class Learner(MessageHandler):
    '''
    This class listens to Accepted messages, determines when the final value is
    selected, and tracks which peers have accepted the final value.
    '''

    def __init__(self, network_uid, quorum_size):
        self.network_uid = network_uid
        self.quorum_size = quorum_size
        self.proposals = dict()  # maps proposal_id => ProposalStatus
        self.acceptors = dict()  # maps from_uid => last_accepted_proposal_id
        self.final_value = None
        self.final_acceptors = None  # Will be a set of acceptor UIDs once the final value is chosen
        self.final_proposal_id = None

    def receive_accepted(self, msg):
        '''
        Called when an Accepted message is received from an acceptor. Once the final value
        is determined, the return value of this method will be a Resolution message containing
        the consentual value. Subsequent calls after the resolution is chosen will continue to add
        new Acceptors to the final_acceptors set and return Resolution messages.
        '''
        if self.final_value is not None:
            if msg.proposal_id >= self.final_proposal_id and msg.proposal_value == self.final_value:
                self.final_acceptors.add(msg.from_uid)
            return Resolution(self.network_uid, self.final_value)

        last_pn = self.acceptors.get(msg.from_uid)

        if last_pn is not None and msg.proposal_id <= last_pn:
            return  # Old message

        self.acceptors[msg.from_uid] = msg.proposal_id

        if last_pn is not None:
            # String proposal_key, need string keys for JSON.
            proposal_key = str(last_pn)
            ps = self.proposals[proposal_key]
            ps.retain_count -= 1
            ps.acceptors.remove(msg.from_uid)
            if ps.retain_count == 0:
                del self.proposals[proposal_key]

        # String proposal_key, need string keys for JSON.
        proposal_key = str(msg.proposal_id)
        if not proposal_key in self.proposals:
            self.proposals[proposal_key] = ProposalStatus(msg.proposal_value)

        ps = self.proposals[proposal_key]

        assert msg.proposal_value == ps.value, 'Value mismatch for single proposal!'

        ps.accept_count += 1
        ps.retain_count += 1
        ps.acceptors.add(msg.from_uid)

        if ps.accept_count == self.quorum_size:
            self.final_proposal_id = msg.proposal_id
            self.final_value = msg.proposal_value
            self.final_acceptors = ps.acceptors
            self.proposals = None
            self.acceptors = None

            return Resolution(self.network_uid, self.final_value)


class PaxosInstance(Proposer, Acceptor, Learner):
    '''
    Aggregate Proposer, Accepter, & Learner class.
    '''

    def __init__(self, network_uid, quorum_size, promised_id=None, accepted_id=None, accepted_value=None):
        Proposer.__init__(self, network_uid, quorum_size)
        Acceptor.__init__(self, network_uid, promised_id, accepted_id, accepted_value)
        Learner.__init__(self, network_uid, quorum_size)

    def receive_prepare(self, msg):
        self.observe_proposal(msg.proposal_id)
        return super(PaxosInstance, self).receive_prepare(msg)

    def receive_accept(self, msg):
        self.observe_proposal(msg.proposal_id)
        return super(PaxosInstance, self).receive_accept(msg)
